import math
import random
import numpy as np

from .pos import Pos
from .constants import DRAWING_SIZE, BLACK, BLUE
from .fractal_constants import DEFAULT_MIN_DIAMETER, DEFAULT_MAX_ITERATIONS
from .helper_fns import interpolate_colour
from .numpy_helper_fns import vect, vect_len, mx_rotd, metric_matrix_min_eig_val, metric_matrix_rms, metric_matrix_x_coord


# -------------------------------------
# General methods

# Turn Numpy vector into Pos
# Pos has origin in top left, vectors in bottom left, so invert in y-axis
def get_canvas_pos_from_vect(vect):
    return Pos(vect[0], DRAWING_SIZE - vect[1])


# Get an interpolated colour using:
# - a list of colours,
# - a progress factor between 0 and 1 representing how far we are through the list
# - an alpha factor saying how much to fade the opacity, between 0 and 1
# - whether to snap to the nearest colour in the list, or interpolate (default)
def get_colour(colours, progress, alpha=1, snap=False):
    max_prog = len(colours) - 1
    prog2 = max(0, min(max_prog, progress * max_prog))
    prog_rem = prog2 - math.floor(prog2)
    colour_start = colours[math.floor(prog2)]
    colour_end = colours[math.ceil(prog2)]
    if snap:
        # Override start and end colours to snap to nearest colour
        colour_start = colours[round(prog2)]
        colour_end = colours[round(prog2)]
    colour_this = interpolate_colour(colour_start, colour_end, prog_rem, alpha)
    return colour_this


# -------------------------------------
# Iteration functions. On a piece, return True to iterate, False to not iterate

# Standard function. Iterate if piece is not too small or too iterated
def get_iteration_fn_standard(min_diameter, max_iterations):
    def iteration_fn(piece):
        return min_diameter < piece.get_minimum_diameter() and piece.iteration < max_iterations

    return iteration_fn


# Turn iteration off
# Use as an override on a definition that should stop iterating
def get_iteration_fn_stop():
    def iteration_fn(piece):
        return False

    return iteration_fn


DEFAULT_ITERATION_FN = get_iteration_fn_standard(DEFAULT_MIN_DIAMETER, DEFAULT_MAX_ITERATIONS)


# -------------------------------------
# Metrics for Fractal Pieces
# Default metric is set on the FractalSystem
# Can override for any particular FractalDefn
# Currently using only the piece's matrix, but could depend on many things,
# such as minimum diameter of convex hull after shear/stretching

# Use the minimum eigenvalue of the matrix
def get_metric_fn_piece_min_eig():
    def metric_fn(piece):
        return metric_matrix_min_eig_val(piece.get_mx())

    return metric_fn


# Use the RMS of the matrix entries
def get_metric_fn_piece_rms():
    def metric_fn(piece):
        return metric_matrix_rms(piece.get_mx())

    return metric_fn


# Use the length of x-coord (1, 0) under mx transformation, e.g. for line fractals
def get_metric_fn_piece_x_coord():
    def metric_fn(piece):
        return metric_matrix_x_coord(piece.get_mx())

    return metric_fn


DEFAULT_METRIC_FN = get_metric_fn_piece_min_eig()


# -------------------------------------
# Random vector functions for drawing-by-hand "hand wobble" effect

# Return 2D square or 3D cube uniform distribution
def wobble_square(pixels=2, dim=2):
    def get_rand_unif():
        return random.uniform(-0.5 * pixels, 0.5 * pixels)

    def wobble_fn():
        x, y, z = get_rand_unif(), get_rand_unif(), get_rand_unif()
        return vect(x, y, z) if dim == 3 else vect(x, y)

    return wobble_fn


# TODO: 2D circle uniform, 3D sphere uniform, 2D concentrated at centre, etc

# Return vector in a rectangular grid. Example (considering x-coord only):
# x_steps = 4
# x_min = -1
# x_max = 1
# Then supply x_pos = 0, 1, 2, 3
# Output is x_this = -0.75, -0.25, 0.25, 0.75
def grid_generator(x_steps, y_steps, x_min=-1, y_min=-1, x_max=1, y_max=1):
    def grid(x_pos, y_pos):
        x_progress = (x_pos + 0.5) / x_steps
        y_progress = (y_pos + 0.5) / y_steps
        x_this = x_min + x_progress * (x_max - x_min)
        y_this = y_min + y_progress * (y_max - y_min)
        return vect(x_this, y_this)

    return grid


# -------------------------------------
# Internal helpers for plotting functions

# Helper function to shrink a single point (q) by a specified brush radius (width)
# How it shrinks depends on the points before and after (p, r)
def shrink_2D_vect(p, q, r, width, mx):
    # Numpy vector arithmetic used here
    # mx ought to be a 90 degree rotation either to left or to right
    # L1 is line from p to q
    # L2 is line from q to r
    # Goal is to shift both L1 and L2 sideways by `width`, then calculate their intersection
    # which is the shrunk coordinate

    # Peturb p, q, r infinitesimally here to avoid potential divisions by zero later on
    delta_v = 0.000001
    def peturb_vector(sc):
        return np.array((sc * (-0.5 + random.random()), sc * (-0.5 + random.random())))
    p = p + peturb_vector(delta_v)
    q = q + peturb_vector(delta_v)
    r = r + peturb_vector(delta_v)

    # Calculate shift vectors to move each line L1, L2 by width, in direction of mx
    sv1_0 = mx @ (q - p)
    sv2_0 = mx @ (r - q)
    sv1 = sv1_0 * (width / vect_len(sv1_0))
    sv2 = sv2_0 * (width / vect_len(sv2_0))

    # Calculate vectors shifted by width here
    ps1 = p + sv1
    qs1 = q + sv1
    qs2 = q + sv2
    rs2 = r + sv2

    # Convert to individual components
    # Now LA is L1 shifted by width, LA is between (x1, y1) and (x2, y2)
    # and LB is L2 shifted by width, LB is between (x3, y3) and (x4, y4)
    x_1 = ps1[0]
    y_1 = ps1[1]
    x_2 = qs1[0]
    y_2 = qs1[1]
    x_3 = qs2[0]
    y_3 = qs2[1]
    x_4 = rs2[0]
    y_4 = rs2[1]

    # Due to x<->y symmetry in line definition, define a function that can obtain either
    def get_coord(x1, x2, x3, x4, y1, y2, y3, y4):
        A = y4 - y3
        B = y2 - y1
        C = y1 - y3
        D = x4 - x3
        E = x2 - x1
        F = (1/(B*D) - 1/(A*E))
        x0 = C/(A*B) + x3/(B*D) - x1/(A*E)
        x = x0 / F
        return x
    
    # Use function to obtain both x and y coords, and return result
    x = get_coord(x_1, x_2, x_3, x_4, y_1, y_2, y_3, y_4)
    y = get_coord(y_1, y_2, y_3, y_4, x_1, x_2, x_3, x_4)
    return np.array((x, y))


# Helper function to shrink 2D path (vect_list) by a specified brush radius (width)
def shrink_2D_vects(piece, vect_list, width):
    result_list = []
    count_vect = len(vect_list)
    # 1. Check we have at least 3 vectors in list to shrink
    if count_vect < 3:
        return vect_list
    # 2. If piece is too small, reduce the shrink amount (width)
    width_from_piece = 0.5 * piece.get_minimum_diameter()
    if width_from_piece < width:
        width = width_from_piece
    # 3. Get the average vector in the list
    av_vect = vect_list[0]
    for i in range(1, count_vect):
        av_vect = av_vect + vect_list[i]
    av_vect = av_vect * (1/count_vect)
    # 4. Check normal (constructed using mx) to line between first two vectors
    # is pointing in same direction as the vector from middle of line to middle of shape
    mx = mx_rotd(angle=90, scale=1)
    p = vect_list[0]
    q = vect_list[1]
    v1 = mx @ (q-p)
    v2 = av_vect * 2 - p - q
    dot_product = sum(v1 * v2)
    if dot_product < 0:
        mx = mx_rotd(angle=-90, scale=1)  # Reverse the orientation
    # 5. Shrink the vector list by specified width, in correct orientation mx
    for i in range(count_vect):
        p = vect_list[(i-1) % count_vect]
        q = vect_list[i]
        r = vect_list[(i+1) % count_vect]
        shrunk_q = shrink_2D_vect(p, q, r, width, mx)
        result_list.append(shrunk_q)
    return result_list


# Helper function to calculate a 2D planar spiral filling path from an outline or convex hull
def spiral_2D_path_fill(vect_list, width):
    n = len(vect_list)  # hull length
    avg_vect = vect_list[0] * 0  # Going to spiral in towards this point
    for i in range(n):
        avg_vect += vect_list[i]
    avg_vect /= n  # turn sum to (mean) average
    max_distance_from_avg = 0
    for i in range(n):
        max_distance_from_avg = max(max_distance_from_avg, vect_len(vect_list[i] - avg_vect))
    # Calculate m, number of spirals in to the centre
    # width is brush radius, so x2
    # Add 2 to ensure overlapping, round to integer number of spirals
    m = round(2 + max_distance_from_avg / (2 * width))
    # Construct list of vectors to plot
    mid_boundary_point = (vect_list[-2] + vect_list[-1]) * 0.5
    spiral_list = [mid_boundary_point, vect_list[-1]] + vect_list  # standard array concatenation
    for i in range(m):
        for j in range(n):
            spiral_list.append(((m - (i + 1)) * vect_list[j] + (i + 1) * avg_vect) * (1 / m))
    spiral_list.append(avg_vect)
    final_draw_list = spiral_list[::-1]  # Reverse order to spiral outwards
    return final_draw_list

# Helper function to do the drawing based on a wide range of criteria
def basic_plot_path(drawing, piece, vector_list, width, shrink, colour, fill, closed, curved, expand_factor, wobble_fn):
    piece_vect = piece.get_vect()
    piece_mx = piece.get_mx()

    # 1. Calculate draw_list, which is the path to draw
    draw_list = []
    # 1.1. Transform vector list to draw list
    for i in range(len(vector_list)):
        wobble_vect = wobble_fn() if callable(wobble_fn) else piece_vect * 0
        draw_vect = piece_vect + wobble_vect + (piece_mx @ vector_list[i]) * expand_factor
        draw_list.append(draw_vect)
    # 1.2. Shrink a bit (due to finite brush radius) if shrink option is selected
    if shrink:
        draw_list = shrink_2D_vects(piece, draw_list, width)
    # 1.3. Deal with both filled and closed shapes
    if fill and len(draw_list) > 2:
        draw_list = spiral_2D_path_fill(vect_list=draw_list, width=width)
        # if filling, shape is automatically closed
    elif closed:
        draw_list.append(draw_list[0])
    
    # 2. Convert to pos_list, a list of Pos objects
    pos_list = []
    for vect in draw_list:
        pos_list.append(get_canvas_pos_from_vect(vect))
    
    # 3. Choose either a straight line or a curved path function to do the drawing
    if curved:
        drawing.add_quadratic_bezier_curve(pos_list, colour, width)
    else:
        drawing.add_line(pos_list, colour, width)


# -------------------------------------
# Plotting functions

# Plot a dot (small circle) for each fractal piece
# Optional parameters:
# expand_factor is to fine-tune control of dot size
# offset_vect is to offset the dot by a specified amount
# wobble_fn returns a random pixel displacement to simulate drawing by hand
def plot_dot(expand_factor=1, wobble_fn=None, offset_vect=None):
    def plot_fn(drawing, piece, colour=BLACK):
        piece_vect = piece.get_vect()
        wobble_vect = wobble_fn() if callable(wobble_fn) else piece_vect * 0
        if not offset_vect is None:
            mx = piece.get_mx()
            piece_vect = piece_vect + mx @ offset_vect
        pos = get_canvas_pos_from_vect(piece_vect + wobble_vect)
        dot_radius = 0.5 * expand_factor * piece.get_minimum_diameter()
        drawing.add_point(pos, colour, dot_radius)

    return plot_fn


# Plot a path (optionally filled, or closed) for each fractal piece separately
# 1. If vector_list is specified, use that for the path
# 2. Else if convex hull is calculated, use that
# 3. Otherwise fall back to a unit square, making sure it is closed
def plot_path(vector_list=None, fill=False, closed=False, curved=False, width=1, shrink=False, expand_factor=1, wobble_fn=None):
    def plot_fn(drawing, piece, colour=BLACK):
        use_closed = closed
        vect_list = vector_list
        if vect_list is None:
            vect_list = piece.get_defn().hull
        if vect_list is None:
            vect_list = [vect(0, 1), vect(-1, 1), vect(-1, -1), vect(1, -1), vect(1, 1)]
            use_closed = True
        basic_plot_path(drawing=drawing, piece=piece, vector_list=vect_list, wobble_fn=wobble_fn, closed=use_closed, colour=colour, width=width, shrink=shrink, curved=curved, expand_factor=expand_factor, fill=fill)

    return plot_fn


DEFAULT_PLOTTING_FN = plot_dot()


# -------------------------------------
# Colouring functions

# Simple colouring function, with 1 fixed colour throughout
# Option exists to vary the alpha, so still going via get_colour instead of
# just returning the fixed colour
def colour_fixed(colour, alpha=1):
    def colour_fn(piece):
        return get_colour([colour], progress=0, alpha=alpha)

    return colour_fn


# Colour by progress, which is a property on each fractal piece
def colour_by_progress(colours, alpha=1, snap=False):
    def colour_fn(piece):
        return get_colour(colours, piece.get_progress_value(), alpha, snap)

    return colour_fn


# Colour by a function of the piece's affine transformation (vector, matrix)
# tsfm(vect, matrix) should output a number
def colour_by_tsfm(min_val, max_val, colours, tsfm, alpha=1, snap=False):
    def colour_fn(piece):
        if not callable(tsfm):
            return BLACK
        this_val = tsfm(piece.get_vect(), piece.get_mx())
        tsfm_progress = (this_val - min_val) / (max_val - min_val)
        return get_colour(colours, tsfm_progress, alpha, snap)

    return colour_fn


# Colour by a function of the piece's affine transformation (vector, matrix)
# metric(matrix) should output a number
def colour_by_log2_size(min_val, max_val, colours, metric=metric_matrix_min_eig_val, alpha=1, snap=False):
    fn = lambda vect, mx: math.log(metric(mx), 2)
    return colour_by_tsfm(min_val, max_val, colours, fn, alpha, snap)


DEFAULT_COLOURING_FN = colour_by_progress([BLACK, BLUE])


# -------------------------------------
# Sorting functions for lists of fractal pieces
# Usage:
# fractal_system.piece_sorter = sort_function(arguments_if_needed)

# Sort pieces randomly
def sort_randomly():
    def sort_fn(piece):
        return random.random()

    return sort_fn


# Sort by function of the piece's affine transformation (vector, matrix)
# Random factor is optional
def sort_by_tsfm(tsfm, rand=False):
    def sort_fn(piece):
        random_factor, main_factor = 0, tsfm(piece.get_vect(), piece.get_mx())
        if rand:
            random_factor = random.random()
        return main_factor + random_factor

    return sort_fn


# Sort by z-coordinate (reversed), e.g. for 3D fractals
def sort_by_z():
    return sort_by_tsfm(lambda vect, mx: -vect[2])


# Sort by size
def sort_by_size():
    return sort_by_tsfm(lambda vect, mx: -metric_matrix_min_eig_val(mx))



import random

from .fractal_plotter import FractalPlotter
from .fractal_piece import FractalPiece
from .fractal_hull_helper_fns import iterate_defn_hull, calculate_hull_diameter
from .fractal_constants import DEFAULT_HULL_ACCURACY


class FractalDefn:
    def __init__(self, system, fid):
        self.system = system  # Link back to the system
        self.fid = fid  # The definition fractal id (fid) number within the system

        # Fractal system contains default metric and iteration functions.
        # Override here at the definition level if needed.
        self.metric_fn = None
        self.iteration_fn = None

        self.plotter = FractalPlotter()  # used to control plotting of FractalPieces linked to this FractalDefn
        self.hull = None  # Convex Hull; set of coordinates describing shape of definition at vector (0, 0), matrix ((1, 0), (0, 1))
        self.hull_accuracy = DEFAULT_HULL_ACCURACY  # hull coordinates will be combined if they are nearer than this (small) number

        # Children should be either a list of abstract Fractal Pieces, or a function returning a list of abstract Fractal Pieces
        # If a function, it should accept an optional FractalPiece as context for evaluation.
        self.children = []

        # What is the minimum diameter of the convex hull of a piece calculated
        # on this definition with identity transformation (vect(0, 0), mx_id())?
        # (Note that if shear or stretch transformations are used then we really ought to calculate
        # diameter directly on the convex hull of each individual piece.)
        # For now, allow this value to be overridden, since perhaps in some situations the convex hull cannot be calculated.
        self.relative_diameter = 2
        # Default value of 2 means that under some rotation the fractal can be fitted between two planes 2 units apart.
        # or that approximately the fractal fits inside the 2x2 square [-1, 1] x [-1, 1]

    def get_metric_fn(self):
        return self.system.metric_fn if self.metric_fn is None else self.metric_fn

    def should_piece_iterate(self, piece):
        iteration_fn = self.system.iteration_fn if self.iteration_fn is None else self.iteration_fn
        return iteration_fn(piece)

    def get_piece_minimum_diameter(self, piece):
        # An alternative to the current method is calculating it directly
        # on the convex hull from this definition
        # transformed by the transformation on an individual piece.
        # This would cope well with stretches and shears.
        metric_fn = self.get_metric_fn()
        return self.relative_diameter * metric_fn(piece)

    def get_children(self, context_piece=None):
        # If self.children is not a function, it should be a list of FractalPieces, so return this list directly
        if not callable(self.children):
            return self.children
        # Otherwise, self.children is a function that should return a list of FractalPieces,
        # and accept an optional context_piece as context for evaluation
        # (The function should be able to cope with and without the context_piece,
        # since for fractal iteration the piece is supplied, but for convex hull it is not.)
        return self.children(context_piece)

    def create_child(self, fid, vect, mx, reverse_progress=False, reset_progress=False):
        # This will be an "Abstract" fractal piece
        piece = FractalPiece(system=self.system, fid=fid, vect=vect, mx=mx, reverse_progress=reverse_progress, reset_progress=reset_progress)
        if callable(self.children):
            # If self.children is a function, remove the function and replace by list
            # Could alternatively raise an error if try to create_child on a fractal defn with fn for children
            self.children = []
        self.children.append(piece)
        return self

    def count_children(self):
        return 0 if callable(self.children) else len(self.children)

    def shuffle_children(self):
        if not callable(self.children):
            random.shuffle(self.children)
        return self

    def remove_child(self, position=None):
        # If position not specified, will remove the last child
        if not callable(self.children):
            if position is None:
                self.children.pop()
            else:
                self.children.pop(position)
        return self

    def plot(self, drawing, piece):
        self.plotter.plot(drawing, piece)

    def initialise_hull(self, hull_accuracy, initial_hull):
        self.hull_accuracy = hull_accuracy
        self.hull = initial_hull.copy()

    def iterate_hull(self, iteration):
        iterate_defn_hull(system=self.system, defn=self, iteration=iteration)

    def calculate_diameter(self):
        # TODO: Setting self.relative_diameter is only accurate in all cases for translations, rotations, reflections
        # If transformation is stretch or shear, this ought to be recalculated at the piece level!
        calculate_hull_diameter(system=self.system, defn=self)

    def __repr__(self):
        result = "FD: ["
        first = True
        if callable(self.children):
            result += "function"
        else:
            for child in self.children:
                if first:
                    first = False
                else:
                    result += ", "
                result += f"{child}"
        result += "]"
        return result



import numpy as np
from scipy.spatial import ConvexHull

from .numpy_helper_fns import mx_rotd


# To make a convex hull in 2D:
# 1. Start with any 2D shape (e.g. a triangle around the origin) for a hull for each definition
# 2. At each calculation iteration, for each definition, take union of previous hulls under fractal maps
# 3. Scale coordinates up (e.g. multiple by 1/0.05) and round to integer, which will deduplicate points close to each other
# 4. Use a convex hull algorithm to remove all interior points (scipy package used here)
# 5. Replace previous hull points with new hull points, and iterate several times

# Note on 2. - currently "vect" is actually not a column vector, but is a row coordinate
# which mean that premultiplying by matrix -> postmultiplying by matrix transpose
# (all the matrix arithmetic is backwards!)
# This probably ought to be fixed later on...

# Also note that for random fractals the hull may not converge so nicely
# This could be alleviated by using multiple (random) copies of the hull points at each stage
# or in some cases by making random fractals (with no context piece supplied) evaluate in the largest possible way

def iterate_defn_hull(system, defn, iteration):
    children = defn.get_children()
    len_children = len(children)
    if len_children < 1:
        return
    next_points = None
    for child in children:
        # FractalDefn contains abstract FractalPieces. These can have fid, vect, mx all functions.
        # Evaluate without any context_piece. These functions should return the largest possible orientation
        # so that convex hull is too big, rather than too small.
        fid = child.get_fid()
        vect = child.get_vect()
        mx = child.get_mx()
        prev_hull = system.lookup_defn(fid).hull
        next_points_partial = prev_hull @ np.transpose(mx) + vect  # Backwards arithmetic here (*)
        if next_points is None:
            next_points = next_points_partial
        else:
            next_points = np.concatenate((next_points, next_points_partial))
    # Not actually going to find convex hull on these points, but on a rounded and scaled version of them
    # which can significantly reduce the number of points in the hull (good for drawing and calculation speed)
    scaled_integer_points = np.rint(next_points * (1 / defn.hull_accuracy))
    hull = ConvexHull(scaled_integer_points)  # object
    vertices = hull.vertices  # array of point indices
    # Now can do lookup on original points (next_points) rather than the approximate and scaled points (scaled_integer_points)
    defn.hull = next_points[vertices]
    # TODO: fix error if hull accuracy is too big then we don't get enough points to make a convex hull, and the method from scipy breaks
    # system.log(f"Hull iteration {iteration} of definition {defn.fid} has length {len(defn.hull)}")


degrees_30 = mx_rotd(angle=30, scale=1)
degrees_60 = mx_rotd(angle=60, scale=1)


def calculate_hull_diameter(system, defn):
    # Calculates minimum diameter over a total of 6 directions: 0, 30, 60, 90, 120, 150 degrees
    hull = defn.hull

    def get_square_min_diam(the_hull):
        x_min, y_min = 10 ** 10, 10 ** 10
        x_max, y_max = -10 ** 10, -10 ** 10
        for point in the_hull:
            x_min = min(x_min, point[0])
            x_max = max(x_max, point[1])
            y_min = min(y_min, point[0])
            y_max = max(y_max, point[1])
        square_diam = min(abs(x_max - x_min), abs(y_max - y_min))
        return square_diam

    def get_min_diam_30_degrees(the_hull):
        d0 = get_square_min_diam(the_hull)
        d30 = get_square_min_diam(the_hull @ degrees_30)
        d60 = get_square_min_diam(the_hull @ degrees_60)
        return min(d0, d30, d60)

    diam = get_min_diam_30_degrees(hull)
    defn.relative_diameter = diam
    system.log(f"- hull {defn.fid} has {len(hull)} points, {diam:.2f} min diameter")



import math
import os
import random

from PIL import ImageColor

from .pos import Pos
from .constants import DRAWING_SIZE


# General helper functions for drawing in Nifty Ink, not related to Numpy or fractals

# Rotate Pos instance (coord) around either the centre of the image, or a specified origin
def rotate(coord, rotation, origin=None):
    # get default centre
    if origin is None:
        origin = Pos(DRAWING_SIZE / 2, DRAWING_SIZE / 2)

    coord -= origin
    x = coord.x * math.cos(rotation) - coord.y * math.sin(rotation)
    y = coord.x * math.sin(rotation) + coord.y * math.cos(rotation)
    return Pos(x, y) + origin


# Turn colour model Hue-Saturation-Value-Alpha into Red-Green-Blue-Alpha
# Currently RBA is in range 0..255 and A is in range 0..1
def hsva_to_rgba(h, s, v, a=1.0):
    i = math.floor(h * 6)
    f = h * 6 - i
    p = v * (1 - s)
    q = v * (1 - f * s)
    t = v * (1 - (1 - f) * s)

    if i % 6 == 0:
        r, g, b = v, t, p
    elif i % 6 == 1:
        r, g, b = q, v, p
    elif i % 6 == 2:
        r, g, b = p, v, t
    elif i % 6 == 3:
        r, g, b = p, q, v
    elif i % 6 == 4:
        r, g, b = t, p, v
    elif i % 6 == 5:
        r, g, b = v, p, q
    else:
        r, g, b = v, t, p

    return [r * 255, g * 255, b * 255, a]


# Convert a hex colour code to an rgba colour list
# e.g. "#FFFFFF" to [255, 255, 255, 1]
def hex_to_rgba(hex_code, alpha=1):
    col_rgb = ImageColor.getcolor(hex_code, "RGB")  # RGB tuple
    col_rgba = list(col_rgb) + [alpha]  # RGBA list
    return col_rgba


# Convert a list of hex colour codes to a list of rgba colour lists
def hex_list_to_rgba(hex_list, alpha=1):
    col_list = []
    for hex_code in hex_list:
        col_list.append(hex_to_rgba(hex_code, alpha))  # List of RGBA lists
    return col_list


# Obtained from https://orthallelous.wordpress.com/2020/06/21/pure-python-bezier-curve/
# TODO: Change this over to using Pos instead of (x, y). Might be quite difficult
def get_bezier_curve(control_points, step_size=10, end_point=True):
    m, q, bezier_points, s = list(zip(*control_points)), len(control_points), [], (
        step_size - 1 if end_point else step_size) / 1.
    for i in range(step_size):
        b = [math.comb(q - 1, v) * (i / s) ** v * (1 - (i / s)) ** (q - 1 - v) for v in range(q)]
        bezier_points += [(tuple(sum(j * k for j, k in zip(d, b)) for d in m))]
    return bezier_points


# Range-check any number into an integer within specified upper and lower bounds
def get_bounded_int(lowest_integer, highest_integer, num):
    return max(lowest_integer, min(highest_integer, round(num)))


# Interpolate between two RGBA values, with optional fading (alpha_factor < 1)
# (R, G, B, A) arrays should be supplied for start and end colours
# RGB in range 0..255, A in range 0..1
def interpolate_colour(start_col, end_col, amount, alpha_factor=1):
    r = get_bounded_int(0, 255, start_col[0] * (1 - amount) + end_col[0] * amount)
    g = get_bounded_int(0, 255, start_col[1] * (1 - amount) + end_col[1] * amount)
    b = get_bounded_int(0, 255, start_col[2] * (1 - amount) + end_col[2] * amount)
    a = max(0, min(1, alpha_factor * (start_col[3] * (1 - amount) + end_col[3] * amount)))
    return r, g, b, a


def alpha_blend(a, bg, fg):
    return ((1 - a) * fg[0] + a * bg[0],
            (1 - a) * fg[1] + a * bg[1],
            (1 - a) * fg[2] + a * bg[2],
            255)


def random_seed():
    seed = int.from_bytes(os.urandom(8), byteorder="big")
    random.seed(seed)
    return seed


def set_random_seed(seed):
    random.seed(seed)



import json
import math
import random

from .pos import Pos
from .font import Font
from .helper_fns import get_bezier_curve, rotate
from .constants import DRAWING_SIZE, DEFAULT_BRUSH_RADIUS, MIN_BRUSH_RADIUS, BLACK, RED


# The Drawing class contains all the code required to produce an output.txt file.
# Copy and paste contents of output.txt code (Javascript) into the web browser on the Create mode of Nifty Ink,
# and it will draw whatever you have coded onto the Nifty Ink canvas.
# Options exist to either overwrite existing canvas, or add a layer on top.

class Drawing:
    def __init__(self):
        self.object = {"lines": [],
                       "width": DRAWING_SIZE,
                       "height": DRAWING_SIZE}

    # Create a round dot / point at the desired location
    def add_point(self, pos, colour, brush_radius):
        line = {"points": [pos, pos],
                "brushColor": "rgba({},{},{},{})".format(*colour),
                "brushRadius": brush_radius}
        self.object["lines"].append(line)

    # Use a large dot to colour the whole canvas
    def add_background(self, colour):
        # 0.71 is approx square root of 0.5
        self.add_point(Pos(DRAWING_SIZE / 2, DRAWING_SIZE / 2), colour, DRAWING_SIZE * 0.71)

    # Add a gradient between two points
    def add_gradient(self, pos1, pos2, colour1, colour2, divisions=30):
        r_step = (colour2[0] - colour1[0]) / divisions
        g_step = (colour2[1] - colour1[1]) / divisions
        b_step = (colour2[2] - colour1[2]) / divisions
        a_step = (colour2[3] - colour1[3]) / divisions
        size_step = (pos2.y - pos1.y) / divisions
        brush_radius = ((pos2.y - pos1.y) / (divisions - 1))

        for i in range(divisions + 1):
            colour = (colour1[0] + (r_step * i),
                      colour1[1] + (g_step * i),
                      colour1[2] + (b_step * i),
                      colour1[3] + (a_step * i))
            line_pos1 = Pos(pos1.x, pos1.y + (size_step * i))
            line_pos2 = Pos(pos2.x, pos1.y + (size_step * i))
            self.add_straight_line(line_pos1, line_pos2, colour, brush_radius)

    # Add a straight line between two positions on the canvas
    def add_straight_line(self, pos1, pos2, colour, brush_radius):
        line = {"points": [pos1, pos2],
                "brushColor": "rgba({},{},{},{})".format(*colour),
                "brushRadius": brush_radius}
        self.object["lines"].append(line)

    # Add a curved line between a list of points (Pos) on the canvas
    # Note - this line is curved on Nifty Ink
    def add_quadratic_bezier_curve(self, pos_list, colour, brush_radius, enclosed_path=False):
        points_list = []
        for pos in pos_list:
            points_list.append(pos)
        if enclosed_path:
            points_list.append(pos_list[0])
        line = {"points": points_list,
                "brushColor": "rgba({},{},{},{})".format(*colour),
                "brushRadius": brush_radius}
        self.object["lines"].append(line)

    # This function is only really useful for fonts. TrueTypeFonts have compressed bezier curves.
    # These curves are not in the same form as the midpoint based bezier curves of Nifty.ink
    # This function serves as a "close enough" approximation of converting them to nifty bezier curves, without the slow
    # and wasteful iterative approach which would make fonts draw far slower and not be consistently smooth with curves.
    def add_modified_quadratic_bezier_curve(self, pos_list, colour, brush_radius, enclosed_path=False):
        max_index = len(pos_list)-1
        new_pos_list = []

        for index, point in enumerate(pos_list):
            if index == 0:
                new_pos_list.append(point)
            if index < max_index:
                next_point = pos_list[index+1]
                midpoint = Pos((next_point.x+point.x)/2, (next_point.y+point.y)/2)
                new_pos_list.append(midpoint)
            else:
                # This is the last point in the segment
                if enclosed_path:
                    next_point = pos_list[0]
                    midpoint = Pos((next_point.x + point.x) / 2, (next_point.y + point.y) / 2)
                    new_pos_list.append(midpoint)
                    new_pos_list.append(next_point)
                else:
                    new_pos_list.append(point)
        self.add_quadratic_bezier_curve(new_pos_list, colour, brush_radius, enclosed_path=False)

    # Add a series of straight line segments between a list of points (Pos) on the canvas
    def add_line(self, pos_list, colour, brush_radius, enclosed_path=False):
        if pos_list:
            points_list = [pos_list[0]]
            for pos in pos_list[1:-1]:
                points_list.append(pos)
                points_list.append(pos)
            points_list.append(pos_list[-1])
            if enclosed_path:
                points_list.append(pos_list[-1])
                points_list.append(pos_list[0])
            line = {"points": points_list,
                    "brushColor": "rgba({},{},{},{})".format(*colour),
                    "brushRadius": brush_radius}
            self.object["lines"].append(line)

    # Add a bezier curve that is quadratic if you give 3 points, cubic if you give 4 points and so on.
    def add_general_bezier_curve(self, control_points, colour, brush_radius, step_size=40, enclosed_path=False):
        if step_size < 2:
            step_size = 2
        tuple_control_points = [(point.x, point.y) for point in control_points]
        points = get_bezier_curve(tuple_control_points, step_size, end_point=True)
        pos_points = [Pos(point[0], point[1]) for point in points]
        self.add_line(pos_points, colour, brush_radius, enclosed_path=enclosed_path)

    # Add a pause to the canvas, using a point off the canvas
    def add_pause(self, length):
        point = Pos(-10, -10)
        line = {"points": [point for _ in range(length)],
                "brushColor": "rgba(0, 0, 0, 0)",
                "brushRadius": 0}
        self.object["lines"].append(line)

    # Add a square to the canvas
    # The degree of roundedness of the corners is determined by the brush radius
    def add_rounded_square(self, centre_pos, width, colour, brush_radius=DEFAULT_BRUSH_RADIUS, filled=True):
        self.add_rounded_rectangle(centre_pos, width, width, colour, brush_radius=brush_radius, filled=filled)

    # # Add a rectangle to the canvas
    # # The degree of roundedness of the corners is determined by the brush radius
    def add_rounded_rectangle(self, centre_pos, width, height, colour, brush_radius=DEFAULT_BRUSH_RADIUS, filled=True):
        rotate_rectangle = False
        if width > height:
            width, height = height, width
            rotate_rectangle = True

        # init corners to outer positions
        corners = [Pos(0, 0),
                   Pos(0, height),
                   Pos(width, height),
                   Pos(width, 0)]
        for i, corner in enumerate(corners):
            corners[i] = centre_pos - Pos(width / 2, height / 2) + corner

        # Checked (internal) brush radius should be larger than minimum in defaults,
        # and then smaller than half of the square width
        br2 = min(width / 2, max(MIN_BRUSH_RADIUS, brush_radius))

        # calculate line_count, this must be a whole number.
        line_count = round(width / (2 * br2))

        # calculate inner corners from outer and brush
        offsets = [Pos(br2, br2),
                   Pos(br2, -br2),
                   Pos(-br2, -br2),
                   Pos(-br2, br2)]
        for i, corner in enumerate(corners):
            corners[i] = corner + offsets[i]

        # do outer line
        result_corners = corners + [corners[0]]

        if filled:
            # do filling lines. line_step is approximately 2 * br2
            line_step = (width - br2 * 2) / line_count
            for i in range(1, line_count):
                result_corners.append(corners[0].copy() + Pos(line_step * i, 0) + Pos(0, br2))
                result_corners.append(corners[1].copy() + Pos(line_step * i, 0) + Pos(0, -br2))

        if rotate_rectangle:
            result_corners = [rotate(point, math.pi/2, centre_pos) for point in result_corners]
        self.add_line(result_corners, colour, br2)

    # Write text onto the canvas using a custom font specified below
    def write(self, font, pos, lines, draw_bounding_box=False):
        line_pos = pos.copy()

        font_size = font.size
        font_weight = font.weight
        line_spacing = font.line_spacing
        colour = font.colour

        y_offset = Pos(0, font_size * line_spacing)

        if not font_weight:
            font_weight = font_size / 60

        def write_char(char_pos, char, char_num):
            x_offset = Pos(font_size * font[ord(char)].character_width, 0)
            current_left_side_bearing = Pos(0, 0)

            # If first character, remove left side bearing
            if char_num == 0:
                current_left_side_bearing = Pos(font[ord(char)].left_side_bearing, 0)
                x_offset -= current_left_side_bearing * font_size

            char_data = font[ord(char)].char_data
            this_char_bounding_box = [((point.copy() - current_left_side_bearing) * font_size) + char_pos for point in font[ord(char)].bounding_box]

            if draw_bounding_box:
                self.add_line(this_char_bounding_box, colour, font_weight, enclosed_path=True)
                full_char_bounding_box = [Pos(0.0, 0.0), Pos((this_char_bounding_box[1].x - char_pos.x) / font_size, 0.0), Pos((this_char_bounding_box[2].x - char_pos.x) / font_size, 1.0), Pos(0.0, 1.0)]
                full_char_bounding_box = [(point.copy() * font_size) + char_pos for point in full_char_bounding_box]
                self.add_line(full_char_bounding_box, RED, font_weight, enclosed_path=True)

            for segment in char_data:
                this_char_segment = [(point.copy() * font_size) + char_pos - (current_left_side_bearing * font_size) for point in segment]
                self.add_modified_quadratic_bezier_curve(this_char_segment, colour, font_weight, enclosed_path=True)

            return char_pos + x_offset

        for line in lines:
            for index, character in enumerate(line):
                pos = write_char(pos.copy(), character, index)
            line_pos = line_pos + y_offset
            pos = line_pos.copy()

    # Randomly reorder the lines
    # Nifty Ink will animate in an interesting random order
    def shuffle_lines(self):
        random.shuffle(self.object["lines"])

    # Use this to make your drawings slightly less precise, but also reduce their size a lot
    # This can be useful if browser local storage limits start affecting your large drawings
    def __round__(self, n_digits=None):
        for line in self:
            r, g, b, a = [float(x) for x in line['brushColor'][5:-1].split(",")]
            if a == 1:
                line['brushColor'] = "rgb(" + ",".join([str(round(r)), str(round(g)), str(round(b))]) + ")"
            else:
                line['brushColor'] = "rgba(" + ",".join([str(round(r)), str(round(g)), str(round(b)), str(a)]) + ")"

            line['brushRadius'] = round(line['brushRadius'])
            line['points'] = [round(point, n_digits) for point in line['points']]

    # Reverse the drawing order of all the lines, this will mess up the final appearance if lines overlap!
    def __reversed__(self):
        self.object['lines'] = list(reversed(self.object['lines']))
        return self

    # Shrink or expand all the stored lines using multiplication
    def __mul__(self, shrink_size):
        # origin for shrinking
        origin = Pos(DRAWING_SIZE / 2, DRAWING_SIZE / 2)

        # shrink each point
        for line in self:
            for point_index in range(len(line["points"])):
                point_pos = line["points"][point_index]
                point_pos -= origin
                point_pos *= shrink_size
                point_pos += origin
                line["points"][point_index] = point_pos
        return self

    # Adds a specified drawing as a new layer on top of this drawing
    # TODO: Handle canvas size scaling (which we currently don't change anyway)
    def __add__(self, drawing):
        self.object['lines'].extend(drawing.object['lines'])

    def __iter__(self):
        for line in self.object['lines']:
            yield line

    def __len__(self):
        return len(self.object['lines'])

    # Save raw drawing data to a file for later use.
    def export_raw_data(self, file_name, indent=4):
        with open(file_name, "w") as file:
            json.dump(self.to_nifty_object(), file, indent=indent)

    # Load a raw data file and replace the contents of this drawing.
    def import_raw_data(self, file_name):
        with open(file_name, "r") as file:
            self.object = self.from_nifty_object(json.load(file))
        return self

    def to_nifty_object(self):
        temp = []
        for line in self:
            temp.append({
                "points": [pos.point() for pos in line['points']],
                "brushColor": line['brushColor'],
                "brushRadius": line['brushRadius']
            })
        return {
            "lines": temp,
            "width": DRAWING_SIZE,
            "height": DRAWING_SIZE
        }

    @staticmethod
    def from_nifty_object(self, json_dict):
        temp = []
        for line in json_dict['lines']:
            temp.append({
                "points": [Pos(point["x"], point["y"]) for point in line['points']],
                "brushColor": line['brushColor'],
                "brushRadius": line['brushRadius']
            })
        return {
            "lines": temp,
            "width": json_dict["width"],
            "height": json_dict["height"]
        }

    # Nifty import method 1 - deprecated
    def to_nifty_show_import(self):
        return "drawingCanvas.current.loadSaveData(\"" + json.dumps(self.to_nifty_object()).replace('"', '\\"') + "\", false)"

    # Nifty import method 2 - overwrite canvas
    def to_nifty_import(self):
        # Use a minified LZString
        # https://raw.githubusercontent.com/pieroxy/lz-string/master/libs/lz-string.min.js
        lz_string = """var LZString=function(){function o(o,r){if(!t[o]){t[o]={};for(var n=0;n<o.length;n++)t[o][o.charAt(n)]=n}return t[o][r]}var r=String.fromCharCode,n="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=",e="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+-$",t={},i={compressToBase64:function(o){if(null==o)return"";var r=i._compress(o,6,function(o){return n.charAt(o)});switch(r.length%4){default:case 0:return r;case 1:return r+"===";case 2:return r+"==";case 3:return r+"="}},decompressFromBase64:function(r){return null==r?"":""==r?null:i._decompress(r.length,32,function(e){return o(n,r.charAt(e))})},compressToUTF16:function(o){return null==o?"":i._compress(o,15,function(o){return r(o+32)})+" "},decompressFromUTF16:function(o){return null==o?"":""==o?null:i._decompress(o.length,16384,function(r){return o.charCodeAt(r)-32})},compressToUint8Array:function(o){for(var r=i.compress(o),n=new Uint8Array(2*r.length),e=0,t=r.length;t>e;e++){var s=r.charCodeAt(e);n[2*e]=s>>>8,n[2*e+1]=s%256}return n},decompressFromUint8Array:function(o){if(null===o||void 0===o)return i.decompress(o);for(var n=new Array(o.length/2),e=0,t=n.length;t>e;e++)n[e]=256*o[2*e]+o[2*e+1];var s=[];return n.forEach(function(o){s.push(r(o))}),i.decompress(s.join(""))},compressToEncodedURIComponent:function(o){return null==o?"":i._compress(o,6,function(o){return e.charAt(o)})},decompressFromEncodedURIComponent:function(r){return null==r?"":""==r?null:(r=r.replace(/ /g,"+"),i._decompress(r.length,32,function(n){return o(e,r.charAt(n))}))},compress:function(o){return i._compress(o,16,function(o){return r(o)})},_compress:function(o,r,n){if(null==o)return"";var e,t,i,s={},p={},u="",c="",a="",l=2,f=3,h=2,d=[],m=0,v=0;for(i=0;i<o.length;i+=1)if(u=o.charAt(i),Object.prototype.hasOwnProperty.call(s,u)||(s[u]=f++,p[u]=!0),c=a+u,Object.prototype.hasOwnProperty.call(s,c))a=c;else{if(Object.prototype.hasOwnProperty.call(p,a)){if(a.charCodeAt(0)<256){for(e=0;h>e;e++)m<<=1,v==r-1?(v=0,d.push(n(m)),m=0):v++;for(t=a.charCodeAt(0),e=0;8>e;e++)m=m<<1|1&t,v==r-1?(v=0,d.push(n(m)),m=0):v++,t>>=1}else{for(t=1,e=0;h>e;e++)m=m<<1|t,v==r-1?(v=0,d.push(n(m)),m=0):v++,t=0;for(t=a.charCodeAt(0),e=0;16>e;e++)m=m<<1|1&t,v==r-1?(v=0,d.push(n(m)),m=0):v++,t>>=1}l--,0==l&&(l=Math.pow(2,h),h++),delete p[a]}else for(t=s[a],e=0;h>e;e++)m=m<<1|1&t,v==r-1?(v=0,d.push(n(m)),m=0):v++,t>>=1;l--,0==l&&(l=Math.pow(2,h),h++),s[c]=f++,a=String(u)}if(""!==a){if(Object.prototype.hasOwnProperty.call(p,a)){if(a.charCodeAt(0)<256){for(e=0;h>e;e++)m<<=1,v==r-1?(v=0,d.push(n(m)),m=0):v++;for(t=a.charCodeAt(0),e=0;8>e;e++)m=m<<1|1&t,v==r-1?(v=0,d.push(n(m)),m=0):v++,t>>=1}else{for(t=1,e=0;h>e;e++)m=m<<1|t,v==r-1?(v=0,d.push(n(m)),m=0):v++,t=0;for(t=a.charCodeAt(0),e=0;16>e;e++)m=m<<1|1&t,v==r-1?(v=0,d.push(n(m)),m=0):v++,t>>=1}l--,0==l&&(l=Math.pow(2,h),h++),delete p[a]}else for(t=s[a],e=0;h>e;e++)m=m<<1|1&t,v==r-1?(v=0,d.push(n(m)),m=0):v++,t>>=1;l--,0==l&&(l=Math.pow(2,h),h++)}for(t=2,e=0;h>e;e++)m=m<<1|1&t,v==r-1?(v=0,d.push(n(m)),m=0):v++,t>>=1;for(;;){if(m<<=1,v==r-1){d.push(n(m));break}v++}return d.join("")},decompress:function(o){return null==o?"":""==o?null:i._decompress(o.length,32768,function(r){return o.charCodeAt(r)})},_decompress:function(o,n,e){var t,i,s,p,u,c,a,l,f=[],h=4,d=4,m=3,v="",w=[],A={val:e(0),position:n,index:1};for(i=0;3>i;i+=1)f[i]=i;for(p=0,c=Math.pow(2,2),a=1;a!=c;)u=A.val&A.position,A.position>>=1,0==A.position&&(A.position=n,A.val=e(A.index++)),p|=(u>0?1:0)*a,a<<=1;switch(t=p){case 0:for(p=0,c=Math.pow(2,8),a=1;a!=c;)u=A.val&A.position,A.position>>=1,0==A.position&&(A.position=n,A.val=e(A.index++)),p|=(u>0?1:0)*a,a<<=1;l=r(p);break;case 1:for(p=0,c=Math.pow(2,16),a=1;a!=c;)u=A.val&A.position,A.position>>=1,0==A.position&&(A.position=n,A.val=e(A.index++)),p|=(u>0?1:0)*a,a<<=1;l=r(p);break;case 2:return""}for(f[3]=l,s=l,w.push(l);;){if(A.index>o)return"";for(p=0,c=Math.pow(2,m),a=1;a!=c;)u=A.val&A.position,A.position>>=1,0==A.position&&(A.position=n,A.val=e(A.index++)),p|=(u>0?1:0)*a,a<<=1;switch(l=p){case 0:for(p=0,c=Math.pow(2,8),a=1;a!=c;)u=A.val&A.position,A.position>>=1,0==A.position&&(A.position=n,A.val=e(A.index++)),p|=(u>0?1:0)*a,a<<=1;f[d++]=r(p),l=d-1,h--;break;case 1:for(p=0,c=Math.pow(2,16),a=1;a!=c;)u=A.val&A.position,A.position>>=1,0==A.position&&(A.position=n,A.val=e(A.index++)),p|=(u>0?1:0)*a,a<<=1;f[d++]=r(p),l=d-1,h--;break;case 2:return w.join("")}if(0==h&&(h=Math.pow(2,m),m++),f[l])v=f[l];else{if(l!==d)return null;v=s+s.charAt(0)}w.push(v),f[d++]=s+v.charAt(0),h--,s=v,0==h&&(h=Math.pow(2,m),m++)}}};return i}();"function"==typeof define&&define.amd?define(function(){return LZString}):"undefined"!=typeof module&&null!=module&&(module.exports=LZString);"""

        # Set up the json string
        json_string = "var json_string = \"" + json.dumps(self.to_nifty_object()).replace("\"", "\\\"").replace(" ", "") + "\";"

        # Update the session storage with the escaped unicode point compressed json string
        local_storage = """window.localStorage.setItem("drawing", JSON.stringify(LZString.compress(json_string)));"""

        # Refresh the Create Ink page to show the new Canvas Ink
        refresh_page = "location.reload();"

        return lz_string + json_string + local_storage + refresh_page

    # Nifty import method 3 - add layer on top of canvas
    def to_nifty_add_layer_import(self):
        # Use a minified LZString
        # https://raw.githubusercontent.com/pieroxy/lz-string/master/libs/lz-string.min.js
        lz_string = """var LZString=function(){function o(o,r){if(!t[o]){t[o]={};for(var n=0;n<o.length;n++)t[o][o.charAt(n)]=n}return t[o][r]}var r=String.fromCharCode,n="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=",e="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+-$",t={},i={compressToBase64:function(o){if(null==o)return"";var r=i._compress(o,6,function(o){return n.charAt(o)});switch(r.length%4){default:case 0:return r;case 1:return r+"===";case 2:return r+"==";case 3:return r+"="}},decompressFromBase64:function(r){return null==r?"":""==r?null:i._decompress(r.length,32,function(e){return o(n,r.charAt(e))})},compressToUTF16:function(o){return null==o?"":i._compress(o,15,function(o){return r(o+32)})+" "},decompressFromUTF16:function(o){return null==o?"":""==o?null:i._decompress(o.length,16384,function(r){return o.charCodeAt(r)-32})},compressToUint8Array:function(o){for(var r=i.compress(o),n=new Uint8Array(2*r.length),e=0,t=r.length;t>e;e++){var s=r.charCodeAt(e);n[2*e]=s>>>8,n[2*e+1]=s%256}return n},decompressFromUint8Array:function(o){if(null===o||void 0===o)return i.decompress(o);for(var n=new Array(o.length/2),e=0,t=n.length;t>e;e++)n[e]=256*o[2*e]+o[2*e+1];var s=[];return n.forEach(function(o){s.push(r(o))}),i.decompress(s.join(""))},compressToEncodedURIComponent:function(o){return null==o?"":i._compress(o,6,function(o){return e.charAt(o)})},decompressFromEncodedURIComponent:function(r){return null==r?"":""==r?null:(r=r.replace(/ /g,"+"),i._decompress(r.length,32,function(n){return o(e,r.charAt(n))}))},compress:function(o){return i._compress(o,16,function(o){return r(o)})},_compress:function(o,r,n){if(null==o)return"";var e,t,i,s={},p={},u="",c="",a="",l=2,f=3,h=2,d=[],m=0,v=0;for(i=0;i<o.length;i+=1)if(u=o.charAt(i),Object.prototype.hasOwnProperty.call(s,u)||(s[u]=f++,p[u]=!0),c=a+u,Object.prototype.hasOwnProperty.call(s,c))a=c;else{if(Object.prototype.hasOwnProperty.call(p,a)){if(a.charCodeAt(0)<256){for(e=0;h>e;e++)m<<=1,v==r-1?(v=0,d.push(n(m)),m=0):v++;for(t=a.charCodeAt(0),e=0;8>e;e++)m=m<<1|1&t,v==r-1?(v=0,d.push(n(m)),m=0):v++,t>>=1}else{for(t=1,e=0;h>e;e++)m=m<<1|t,v==r-1?(v=0,d.push(n(m)),m=0):v++,t=0;for(t=a.charCodeAt(0),e=0;16>e;e++)m=m<<1|1&t,v==r-1?(v=0,d.push(n(m)),m=0):v++,t>>=1}l--,0==l&&(l=Math.pow(2,h),h++),delete p[a]}else for(t=s[a],e=0;h>e;e++)m=m<<1|1&t,v==r-1?(v=0,d.push(n(m)),m=0):v++,t>>=1;l--,0==l&&(l=Math.pow(2,h),h++),s[c]=f++,a=String(u)}if(""!==a){if(Object.prototype.hasOwnProperty.call(p,a)){if(a.charCodeAt(0)<256){for(e=0;h>e;e++)m<<=1,v==r-1?(v=0,d.push(n(m)),m=0):v++;for(t=a.charCodeAt(0),e=0;8>e;e++)m=m<<1|1&t,v==r-1?(v=0,d.push(n(m)),m=0):v++,t>>=1}else{for(t=1,e=0;h>e;e++)m=m<<1|t,v==r-1?(v=0,d.push(n(m)),m=0):v++,t=0;for(t=a.charCodeAt(0),e=0;16>e;e++)m=m<<1|1&t,v==r-1?(v=0,d.push(n(m)),m=0):v++,t>>=1}l--,0==l&&(l=Math.pow(2,h),h++),delete p[a]}else for(t=s[a],e=0;h>e;e++)m=m<<1|1&t,v==r-1?(v=0,d.push(n(m)),m=0):v++,t>>=1;l--,0==l&&(l=Math.pow(2,h),h++)}for(t=2,e=0;h>e;e++)m=m<<1|1&t,v==r-1?(v=0,d.push(n(m)),m=0):v++,t>>=1;for(;;){if(m<<=1,v==r-1){d.push(n(m));break}v++}return d.join("")},decompress:function(o){return null==o?"":""==o?null:i._decompress(o.length,32768,function(r){return o.charCodeAt(r)})},_decompress:function(o,n,e){var t,i,s,p,u,c,a,l,f=[],h=4,d=4,m=3,v="",w=[],A={val:e(0),position:n,index:1};for(i=0;3>i;i+=1)f[i]=i;for(p=0,c=Math.pow(2,2),a=1;a!=c;)u=A.val&A.position,A.position>>=1,0==A.position&&(A.position=n,A.val=e(A.index++)),p|=(u>0?1:0)*a,a<<=1;switch(t=p){case 0:for(p=0,c=Math.pow(2,8),a=1;a!=c;)u=A.val&A.position,A.position>>=1,0==A.position&&(A.position=n,A.val=e(A.index++)),p|=(u>0?1:0)*a,a<<=1;l=r(p);break;case 1:for(p=0,c=Math.pow(2,16),a=1;a!=c;)u=A.val&A.position,A.position>>=1,0==A.position&&(A.position=n,A.val=e(A.index++)),p|=(u>0?1:0)*a,a<<=1;l=r(p);break;case 2:return""}for(f[3]=l,s=l,w.push(l);;){if(A.index>o)return"";for(p=0,c=Math.pow(2,m),a=1;a!=c;)u=A.val&A.position,A.position>>=1,0==A.position&&(A.position=n,A.val=e(A.index++)),p|=(u>0?1:0)*a,a<<=1;switch(l=p){case 0:for(p=0,c=Math.pow(2,8),a=1;a!=c;)u=A.val&A.position,A.position>>=1,0==A.position&&(A.position=n,A.val=e(A.index++)),p|=(u>0?1:0)*a,a<<=1;f[d++]=r(p),l=d-1,h--;break;case 1:for(p=0,c=Math.pow(2,16),a=1;a!=c;)u=A.val&A.position,A.position>>=1,0==A.position&&(A.position=n,A.val=e(A.index++)),p|=(u>0?1:0)*a,a<<=1;f[d++]=r(p),l=d-1,h--;break;case 2:return w.join("")}if(0==h&&(h=Math.pow(2,m),m++),f[l])v=f[l];else{if(l!==d)return null;v=s+s.charAt(0)}w.push(v),f[d++]=s+v.charAt(0),h--,s=v,0==h&&(h=Math.pow(2,m),m++)}}};return i}();"function"==typeof define&&define.amd?define(function(){return LZString}):"undefined"!=typeof module&&null!=module&&(module.exports=LZString);"""

        # Raw json string
        raw_json_string = json.dumps(self.to_nifty_object()).replace("\"", "\\\"").replace(" ", "")

        # Set up the json string of the new data
        json_object = "var json_object = JSON.parse(\"" + raw_json_string + "\");"

        # Save the canvas if there are any unsaved changes
        save_canvas = """var button = document.getElementsByTagName("button");
                         for (var i = 0; i < button.length; i++) {
                           if (button[i].innerHTML.includes("SAVE *")) {
                             button[i].click();
                           }
                         };"""

        # Get the current saved data from the localStorage
        local_storage_data = """var raw_storage = window.localStorage.getItem("drawing");"""

        # Scale all the points of the newly imported data to the current saved data canvas size
        update_data = """if (raw_storage == "" || raw_storage == "undefined") {
                            var json_string = JSON.stringify(json_object)
                         } else {
                            var decompressed_data = LZString.decompress(JSON.parse(raw_storage));
                            decompressed_data = JSON.parse(decompressed_data);
                            var saved_height = decompressed_data['height'];
                            var saved_width = decompressed_data['width'];
                            var imported_height = json_object['height'];
                            var imported_width = json_object['width'];

                            var scale_height = saved_height / imported_height;
                            var scale_width = saved_width / imported_width;
                            var scale_brush_radius = Math.max(scale_height, scale_width)

                            json_object['lines'].forEach(function(part, i) {
                              this[i]['brushRadius'] = this[i]['brushRadius']*scale_brush_radius;
                              this[i]['points'].forEach(function(part, j) {
                                this[j]['x'] = this[j]['x']*scale_width;
                                this[j]['y'] = this[j]['y']*scale_height;
                              }, this[i]['points']);
                            }, json_object['lines']);

                            decompressed_data['lines'] = decompressed_data['lines'].concat(json_object['lines']);
                            var json_string = JSON.stringify(decompressed_data);
                         };"""

        # Update the session storage with the escaped unicode point compressed json string
        local_storage = """window.localStorage.setItem("drawing", JSON.stringify(LZString.compress(json_string)));"""

        # Refresh the Create Ink page to show the new Canvas Ink Layer
        refresh_page = "location.reload();"

        return lz_string + json_object + save_canvas + local_storage_data + update_data + local_storage + refresh_page



from .constants import BLACK
from .fractal_helper_fns import DEFAULT_PLOTTING_FN, DEFAULT_COLOURING_FN

# A FractalPlotter object controls how a FractalPiece is plotted onto the canvas
# It contains a list of instructions for the plot method and the colour to plot

class FractalPlotter:
    def __init__(self):
        # Set this to false to prevent any drawing from taking place
        self.draws = True

        # Store a list of things to do when plotting
        # Format is:
        # [[plot_fn_1, colour_fn_1], [plot_fn_2, colour_fn_2], ...]
        # Use the `add` function to add a new pair to this list
        self.plot_list = []
    
    # Add an instruction to use a plot method with a particular colour
    # Each of these are functions that get evaluated in the context of a specific fractal piece
    # Plotting function is mandatory, examples include plotting a dot, or plotting a path (filled/outline, closed/open)
    # Colouring function is optional, if not specified then BLACK will be used.
    def add(self, plotting_fn, colouring_fn=None):
        self.plot_list.append([plotting_fn, colouring_fn])
    
    # Plot a specific fractal piece onto the drawing (canvas),
    # using the plot methods stored in this FractalPlotter
    def plot(self, drawing, piece):
        if not self.draws:
            return
        if len(self.plot_list) == 0:
            # If user has not specified any plotting behaviour, lazy-initialise it here
            self.add(DEFAULT_PLOTTING_FN, DEFAULT_COLOURING_FN)
        # Iterate through the list of plotting instructions
        # and execute each one for the fractal piece
        for plot_and_colour_fns in self.plot_list:
            plot_fn = plot_and_colour_fns[0]
            colour_fn = plot_and_colour_fns[1]
            colour = colour_fn(piece) if callable(colour_fn) else BLACK
            if callable(plot_fn):
                plot_fn(drawing, piece, colour)

    def __repr__(self):
        return f"FP: draws {self.draws}"



import numpy as np

DEFAULT_MAX_DEFNS = 10000  # normally would use a lot fewer than this
DEFAULT_MIN_DIAMETER = 30
DEFAULT_MAX_PIECES = 10000
DEFAULT_MAX_ITERATIONS = 6
BREAK_AFTER_ITERATIONS = 1000  # stop the loop if this number of iterations is reached

DEFAULT_HULL_MAX_ITERATIONS = 20
DEFAULT_HULL_ACCURACY = 0.031415  # Acceptable accuracy for fractal roughly filling interval [-1, 1] in each direction

# This hull ought to have diameter 2 in the diameter-calculating algorithm being used
# Current algorithm is to find max-min x and y at 0, 30, 60 degrees (min of 6 values)
# This hull produces 2 under that algorithm.
DEFAULT_INITIAL_HULL = np.array(((1, 1), (-1, 0.5), (0.5, -1)))



import math

from .constants import DRAWING_SIZE


# This class Pos wraps an x and y coordinate.
# This class could potentially be refactored (and replaced) by numpy vectors,
# however there would be a lot of updates to make.

class Pos:
    def __init__(self, x, y):
        self.x = x
        self.y = y

    @staticmethod
    def from_rotational(radians, radius, origin=None):
        if origin is None:
            origin = Pos(DRAWING_SIZE / 2, DRAWING_SIZE / 2)

        return Pos(math.cos(radians) * radius / 2, math.sin(radians) * radius / 2) + origin

    def point(self):
        return {"x": self.x, "y": self.y}

    def copy(self):
        return Pos(self.x, self.y)

    def rotate(self, rotation, origin=None):
        # get default centre
        if origin is None:
            origin = Pos(DRAWING_SIZE / 2, DRAWING_SIZE / 2)

        self -= origin
        x = self.x * math.cos(rotation) - self.y * math.sin(rotation)
        self.y = self.x * math.sin(rotation) + self.y * math.cos(rotation)
        self.x = x
        return self + origin

    def __add__(self, other):
        result = self.copy()
        result.x += other.x
        result.y += other.y
        return result

    def __sub__(self, other):
        result = self.copy()
        result.x -= other.x
        result.y -= other.y
        return result

    def __mul__(self, mult):
        result = self.copy()
        result.x *= mult
        result.y *= mult
        return result

    def __truediv__(self, div):
        result = self.copy()
        result.x /= div
        result.y /= div
        return result

    def __round__(self, n_digits=None):
        result = self.copy()
        result.x = round(self.x, n_digits)
        result.y = round(self.y, n_digits)
        return result

    def __repr__(self):
        return f"({self.x}, {self.y})"



from .fractal_defn import FractalDefn
from .fractal_constants import DEFAULT_MAX_PIECES, DEFAULT_MAX_DEFNS, BREAK_AFTER_ITERATIONS
from .fractal_constants import DEFAULT_HULL_MAX_ITERATIONS, DEFAULT_INITIAL_HULL
from .fractal_helper_fns import DEFAULT_METRIC_FN, DEFAULT_ITERATION_FN


# There should only be 1 fractal system created,
# link all fractal definitions and pieces back to this system

class FractalSystem:
    def __init__(self, max_pieces=DEFAULT_MAX_PIECES):
        self.defns = []
        self.max_pieces = max_pieces
        self.max_defns = DEFAULT_MAX_DEFNS
        self.iteration_fn = DEFAULT_ITERATION_FN  # Function to tell a fractal piece if it should iterate or not
        self.metric_fn = DEFAULT_METRIC_FN  # Function mapping from affine transformation (vect, mx) of a piece, to a numeric value. Can override on the definition.
        self.initial_pieces = []  # Set this one to a list of Fractal Pieces
        self.iterated_pieces = []  # Do not set this one, call fs.do_iterations() to generate it automatically
        self.piece_sorter = None
        self.verbose = True  # Set to false to suppress printing output to console

    def lookup_defn(self, fid):
        if isinstance(fid, int) and fid >= 0 and fid < len(self.defns):
            return self.defns[fid]
        else:
            # If fractal id (fid) is not found, fail gracefully by returning None.
            # Should check for None and only take further action if a definition is returned.
            return None

    def add_defn(self, fractal_defn):
        self.defns.append(fractal_defn)
        fractal_defn.system = self  # Force definition to link back to this system
        return self

    def make_defns(self, n):
        if isinstance(n, int) and n > 0 and n <= self.max_defns:
            self.defns = []
            for fid in range(n):
                self.add_defn(FractalDefn(system=self, fid=fid))
        return self

    def log(self, text):
        if self.verbose:
            print(text)

    def calculate_hulls(self, max_iterations=DEFAULT_HULL_MAX_ITERATIONS, hull_accuracy=None, initial_hull=DEFAULT_INITIAL_HULL):
        # 1. Initialise hulls on all definitions
        self.log("")
        self.log("Initialising convex hull calculations")
        for defn in self.defns:
            defn.initialise_hull(hull_accuracy=hull_accuracy, initial_hull=initial_hull)
        # 2. Iteratively calculate all hulls in parallel (necessary since they interact)
        self.log("")
        self.log("Calculating convex hulls")
        for i in range(max_iterations):
            for defn in self.defns:
                defn.iterate_hull(iteration=i)
            self.log(f"- hulls iteration {i + 1}")
        self.log("")
        self.log("Calculating definition minimum diameters")
        for defn in self.defns:
            defn.calculate_diameter()

    def do_iterations(self):
        self.log("")
        self.log(f"Calculating fractal iterations")
        self.iterated_pieces = self.initial_pieces
        iteration_finished = False
        counter = 0
        while not iteration_finished:
            counter += 1
            self.log(f"- iteration {counter} on {len(self.iterated_pieces)} piece{'' if len(self.iterated_pieces) == 1 else 's'}")
            iteration_finished = self.iterate_once()
            if BREAK_AFTER_ITERATIONS <= counter:
                self.log(f"Forced break after {counter} iterations")
                break
        self.log("")

    def iterate_once(self):
        iteration_finished = True  # Set to false if at least one piece was successfully iterated
        exceeded_max_pieces = False  # Set to true if we run out of storage space
        n = len(self.iterated_pieces)  # self.iterated_pieces is this iteration
        collect_next_iteration = []  # collect_next_iteration is next iteration, updated by .iterate function below
        for i in range(n):
            piece_to_iterate = self.iterated_pieces[i]
            if exceeded_max_pieces or self.max_pieces - (n - i) < len(collect_next_iteration):
                # Have run out of space, just keep the same pieces without iterating
                collect_next_iteration.append(piece_to_iterate)
                exceeded_max_pieces = True
                iteration_finished = True
            else:
                was_iterated = piece_to_iterate.iterate(collect_next_iteration)  # returns Boolean, appends to collect_next_iteration
                if was_iterated:
                    iteration_finished = False  # keep going until no piece iterates any further
        if exceeded_max_pieces:
            self.log("Warning - max pieces exceeded")
        self.iterated_pieces = collect_next_iteration
        return iteration_finished

    def final_size(self):
        return len(self.iterated_pieces)

    def plot(self, drawing):
        if callable(self.piece_sorter):
            self.iterated_pieces.sort(key=self.piece_sorter)
        for piece_to_plot in self.iterated_pieces:
            piece_to_plot.plot(drawing)

    def __repr__(self):
        result = "FS: "
        first = True
        defn_fid = 0
        for defn in self.defns:
            if first:
                first = False
            else:
                result += ", "
            result += f"(Fractal ID {defn_fid} contains {defn})"
            defn_fid += 1
        return result



import datetime
import os
import time
import numpy as np
from PIL import Image


# Hide the Pygame support message
os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = str()
import pygame

from .constants import BLACK, WHITE, DRAWING_SIZE, TITLE_BAR_HEIGHT, BORDER_WIDTH
from .helper_fns import get_bezier_curve, alpha_blend



class Renderer:
    def __init__(self, headless=False, pygame_scale=None):
        self.pygame_scale = pygame_scale
        self.headless = headless

        # Set a fake video driver to hide output
        if headless:
            os.environ['SDL_VIDEODRIVER'] = 'dummy'
            # No screen to get the dimensions, just render at normal size
            if pygame_scale is None:
                self.pygame_scale = 1

        # Init pygame
        pygame.init()

        # Reposition and change size of surface to account for title bar
        if not headless:
            info_object = pygame.display.Info()
            smallest_dimension = min(info_object.current_w, info_object.current_h)

            x = round((info_object.current_w - (smallest_dimension - TITLE_BAR_HEIGHT - (BORDER_WIDTH * 2))) / 2)
            y = TITLE_BAR_HEIGHT + BORDER_WIDTH
            os.environ['SDL_VIDEO_WINDOW_POS'] = "%d,%d" % (x, y)

            # Scale the window and drawing to the maximum square size
            if pygame_scale is None:
                self.pygame_scale = (smallest_dimension - TITLE_BAR_HEIGHT - (BORDER_WIDTH * 2)) / DRAWING_SIZE

        # Initialise the window with dimensions
        self.pygame_x = round(DRAWING_SIZE * self.pygame_scale)
        self.pygame_y = round(DRAWING_SIZE * self.pygame_scale)
        self.screen = pygame.display.set_mode((self.pygame_x, self.pygame_y))
        pygame.display.set_caption("Drawing Render")



    # Render the lines to preview in Pygame
    def render(self, drawing, filename="output.png", simulate=False, speed=None,
               allow_transparency=False, fake_transparency=False, proper_line_thickness=False, draw_as_bezier=False,
               step_size=10, save_transparent_bg=False, green_screen_colour=(0, 177, 64, 255), timestamp=False,
               timestamp_format="%Y_%m_%d_%H_%M_%S_%f_"):

        if step_size < 2:
            step_size = 2

        if save_transparent_bg:
            self.screen.fill(green_screen_colour)
        else:
            self.screen.fill(WHITE)
        # self.screen.set_alpha(255)  CHECK NOT USED
        pygame.display.update()  # Show the background, (so the screen isn't black on drawings that are slow to process)

        def draw_line(surface, colour, start_point, end_point, width, end_caps=False):
            if end_caps:
                pygame.draw.circle(surface, colour, start_point, width / 2)
                pygame.draw.circle(surface, colour, end_point, width / 2)
            if start_point == end_point:
                return
            np.seterr(divide='ignore', invalid='ignore')
            vec_start_point = np.array(start_point)
            vec_end_point = np.array(end_point)
            move_point = vec_end_point - vec_start_point
            norm_move = move_point / np.linalg.norm(move_point)

            rotated_vec = np.array((-norm_move[1], norm_move[0])) * width / 2
            start_point_1 = vec_start_point + rotated_vec
            start_point_2 = vec_start_point - rotated_vec
            end_point_1 = vec_end_point + rotated_vec
            end_point_2 = vec_end_point - rotated_vec

            pygame.draw.polygon(surface, colour, [start_point_1, start_point_2, end_point_2, end_point_1], width=0)

        def draw_lines(surface, colour, pts, width, end_caps=False):
            last_point = None
            for pt in pts:
                if last_point:
                    draw_line(surface, colour, last_point, pt, width, end_caps=end_caps)
                last_point = pt

        def get_midpoint(p1, p2):
            x = (p1[0] + p2[0]) / 2
            y = (p1[1] + p2[1]) / 2
            return [x, y]

        def draw_quadratic_bezier_curve_line(surface, colour, pts, width, end_caps=False, step_size=40):
            if pts:
                last_midpoint = pts[0]
                midpoint = last_midpoint
                p2 = last_midpoint

                for i in range(len(pts)):
                    p1 = pts[i]
                    try:
                        p2 = pts[i + 1]

                        midpoint = get_midpoint(p1, p2)
                        # TODO: Write some code to create an appropriate step_size, likely based on the bezier curve length
                        bezier_curve_points = get_bezier_curve((last_midpoint, p1, midpoint), step_size=step_size,
                                                               end_point=True)
                        draw_lines(surface, colour, bezier_curve_points, width, end_caps=end_caps)

                        last_midpoint = midpoint
                    except IndexError:  # Draw the last point as a straight line to finish
                        draw_line(surface, colour, midpoint, p2, width, end_caps=end_caps)

        for line in drawing:
            brush_radius = line["brushRadius"] * self.pygame_scale
            if "rgba" in line["brushColor"]:
                colour = [float(cell) for cell in list(line["brushColor"][5:-1].split(","))]
                colour[3] *= 255
            else:
                colour = [float(cell) for cell in list(line["brushColor"][4:-1].split(","))]
                colour.append(255)

            points = []
            if colour[3] != 255 and allow_transparency:  # If the brushColour is transparent, draw with transparency
                target_surface = pygame.Surface((self.pygame_x, self.pygame_y), 0, 32)
                if colour[:-1] != [0, 0, 0]:
                    target_surface.set_colorkey(BLACK)
                else:  # Handle the black edge case
                    target_surface.set_colorkey(WHITE)
                    target_surface.fill((255, 255, 255, 0))
                target_surface.set_alpha(round(colour[3]))
            else:  # If the brushColour is opaque, draw with no transparency
                if fake_transparency:
                    colour = alpha_blend(colour[3] / 255, colour[:-1], [255, 255, 255])
                target_surface = self.screen

            for index, point in enumerate(line["points"]):
                this_point = (point.x * self.pygame_scale, point.y * self.pygame_scale)
                points.append(this_point)
                if not proper_line_thickness:
                    pygame.draw.circle(target_surface, colour, this_point, int(brush_radius))

            if proper_line_thickness:
                if draw_as_bezier:
                    draw_quadratic_bezier_curve_line(target_surface, colour, points, brush_radius * 2,
                                                     end_caps=True, step_size=step_size)
                else:
                    draw_lines(target_surface, colour, points, brush_radius * 2, end_caps=True)
            else:
                pygame.draw.lines(target_surface, colour, False, points, int(brush_radius * 2))

            # Required for transparency
            if colour[3] != 255 and allow_transparency:
                self.screen.blit(target_surface, (0, 0))

            # Update the drawing line by line to see the drawing process
            if simulate:
                pygame.display.update()
                if speed and speed != 0:
                    time.sleep(speed / 100)

            # Ensure that no events, such as pygame being closed are ignored.
            ev = pygame.event.get()
            for event in ev:
                if event.type == pygame.QUIT:
                    # Exits before the image is finished, does not take screenshot.
                    return

        # update screen to render the final result of the drawing
        pygame.display.update()

        time_str = datetime.datetime.now().strftime(timestamp_format) if timestamp else ""

        pygame.image.save(self.screen, time_str + filename)

        # TODO: Figure out if Pygame has a method to save a surface with a transparent background
        if save_transparent_bg:
            image_string = pygame.image.tostring(self.screen, 'RGBA', False)
            img = Image.frombytes("RGBA", self.screen.get_size(), image_string)

            # https://stackoverflow.com/a/69814643
            def convert_png_transparent(image, dst_file, bg_color=(255, 255, 255)):
                array = np.array(image, dtype=np.ubyte)
                mask = (array[:, :, :3] == bg_color).all(axis=2)
                alpha = np.where(mask, 0, 255)
                array[:, :, -1] = alpha

                Image.fromarray(np.ubyte(array)).save(dst_file, "PNG")

            convert_png_transparent(img, f"{time_str}transparent_{filename}", [*green_screen_colour[:-1]])

        # enter a loop to prevent pygame from ending
        running = True
        while running and not self.headless:
            ev = pygame.event.get()
            for event in ev:
                if event.type == pygame.QUIT:
                    running = False
                    break
            time.sleep(0.2)  # Sleep for a short time. Prevents continual use of CPU.


import random

from .fractal_piece import FractalPiece
from .fractal_helper_fns import grid_generator
from .numpy_helper_fns import vect, vect_len, mx_scale, mx_rotd, mx_refl_X, mx_sq, mx_dh


# -------------------------------------
# Some fractal parameters are either values or functions
# Use this file to store some common functions that return suitable values
# Usually the function takes a fractal piece in as an optional parameter
# to fine-tune the behaviour.

# For a square [-1, 1] x [-1, 1]
# split it into n^2 tiles (nxn)
# and then keep m out of n^2 at random
def gen_children_rand_small_squares(system, fid, m, n):
    grid = grid_generator(x_steps=n, y_steps=n)

    def calc_children(context_piece=None):
        # This function is probabilistic, so context piece is not used. Still need the parameter available!
        children = []
        for x in range(n):
            for y in range(n):
                piece = FractalPiece(system=system, fid=fid, vect=grid(x, y), mx=mx_scale(1 / n))
                children.append(piece)
        return random.sample(children, m)

    return calc_children


# For a square [-1, 1] x [-1, 1]
# split it into n^2 tiles (nxn)
# and keep each tile with probability between p1 and p2
# depending on its distance from vect
# For large pieces, iterate them with probability 1,
# only apply the probabilistic iteration to scales (metrics) below the cutoff.
def gen_children_fade_out(system, fid, n, centre_vect, cutoff_diameter, d1=0, d2=2000, p1=1, p2=0):
    grid = grid_generator(x_steps=n, y_steps=n)

    def calc_children(context_piece=None):
        p = 1  # Default is to include all children, unless context is supplied and size (metric) is below cutoff
        if context_piece is not None:
            if context_piece.get_minimum_diameter() < cutoff_diameter:
                outer_vect = context_piece.get_vect()
                d = vect_len(outer_vect - centre_vect)
                d_progress_1_to_2 = max(0, min(1, (d - d1) / (d2 - d1)))
                p = p1 + d_progress_1_to_2 * (p2 - p1)
        children = []
        for x in range(n):
            for y in range(n):
                if (random.random() < p):
                    piece = FractalPiece(system=system, fid=fid, vect=grid(x, y), mx=mx_scale(1 / n))
                    children.append(piece)
        return children

    return calc_children


# -------------------------------------
# Methods to calculate a random fractal id

# Select an fractal id (fid) at random from a list, equal weights
def gen_fid_rand(list_of_fids):
    def calc_id(context_piece=None):
        return random.choice(list_of_fids)

    return calc_id


# -------------------------------------
# 2D or 3D methods to calculate a random vector

# Calculate a random continuous 2D or 3D vector in (x1, x2) x (y1, y2) [x (z1, z2)]
# Example: gen_vect_rand([1, 2], [3, 4], [5, 6])
def gen_vect_rand(x_range, y_range, z_range=None):
    def calc_vect(context_piece=None):
        x = random.uniform(x_range[0], x_range[1])
        y = random.uniform(y_range[0], y_range[1])
        if z_range is None:
            return vect(x, y)
        z = random.uniform(z_range[0], z_range[1])
        return vect(x, y, z)

    return calc_vect


# -------------------------------------
# 2D methods to calculate a random matrix

# Any rotation or reflection in the circle
def gen_mx_rand_circ(scale=1, reflect=True):
    def calc_mx(context_piece=None):
        mx = mx_rotd(
            angle=random.uniform(0, 360),
            scale=scale
        )
        if reflect and (random.random() < 0.5):
            mx = mx @ mx_refl_X()
        return mx

    return calc_mx


# Any rotation or reflection in a square with a flat edge down
def gen_mx_rand_sq(scale=1, reflect=True):
    max_num = 4
    if reflect:
        max_num = 8

    def calc_mx(context_piece=None):
        return mx_sq(
            num=random.randint(1, max_num),
            scale=scale
        )

    return calc_mx


# Any rotation or reflection in a triangle with a flat edge down
def gen_mx_rand_tri(scale=1, reflect=True):
    max_num = 3
    if reflect:
        max_num = 6

    def calc_mx(context_piece=None):
        return mx_dh(
            sides=3,
            num=random.randint(1, max_num),
            scale=scale
        )

    return calc_mx


# Any rotation or reflection in a <sides>-sided polygon with a flat edge down
def gen_mx_rand_dihedral(sides, scale=1, reflect=True):
    max_num = sides
    if reflect:
        max_num = sides * 2

    def calc_mx(context_piece=None):
        return mx_dh(
            sides=sides,
            num=random.randint(1, max_num),
            scale=scale
        )

    return calc_mx



from . import drawing
from . import pos
from . import helper_fns
from . import numpy_helper_fns
from . import constants
from . import drawing_methods

from . import fractal_runner
from . import fractal_constants
from . import fractal_defn
from . import fractal_generator_fns
from . import fractal_helper_fns
from . import fractal_hull_helper_fns
from . import fractal_piece
from . import fractal_plotter
from . import fractal_system

from . import font

from . import renderer



from .fractal_helper_fns import DEFAULT_ITERATION_FN


# Consider splitting FractalPiece class into concrete and abstract subclasses
#
# Abstract fractal pieces are the ones inside a fractal definition,
# and can have values or functions for fid, vect, matrix
# Must use get_fid, get_vect, get_mx for these.
#
# Concrete fractal pieces can only have values (and not functions) for fid, vect, matrix
# Fractal system should have concrete initial pieces, and concrete iterated pieces
# This is mentioned in comments on the 'iterate' method below.
# For concrete, it would still be encouraged to use the getters,
# even though self.fid, self.vect, self.mx would work.

class FractalPiece:
    def __init__(self, system, fid, vect, mx, iteration=0, progress=None, reverse_progress=False, reset_progress=False):
        self.system = system

        # Keep track of what iteration this piece is on
        self.iteration = iteration

        # Keep track of "progress" through the fractal iteration using an interval [start, end]
        # Default interval is [0, 1]. Interval is a list of [start, end] values of progress.
        # When iterating, this interval should be subdivided and become smaller and smaller within [0, 1].
        self.progress = progress if progress is not None else [0, 1]  # use None above and [0, 1] here to get new list each time

        # In a fractal definition, for a child fractal, set reverse_progress directly to True
        # if you want progress to be flipped when iterating.
        # An example of when this is useful is for a dragon fractal, to get linear colouring.
        self.reverse_progress = reverse_progress

        # In a fractal definition, for a child fractal, set reset_progress directly to True
        # if you want progress to revert back to [0, 1] on this fractal piece
        # For example, if using an outer fractal to generate a lot of inner fractals,
        # might want to reset the progress on each one. This could for example reset the colouring scheme on each inner fractal.
        # Probably don't have reset_progress true for any fractal that iterates to itself.
        self.reset_progress = reset_progress

        # These should be either a value of the specified type, or a function returning suitable value
        # Function should accept an optional FractalPiece as context for evaluation.
        self.fid = fid  # Fractal id (fid) of definition. Should be a non-negative integer 0, 1, 2... representing definition position in fractal system
        self.vect = vect  # Numpy vector (or is it coordinate?)
        self.mx = mx  # Numpy matrix

    def get_progress_value(self):
        return 0.5 * (self.progress[0] + self.progress[1])

    # Return array of n intervals representing progress interval of this piece split into n chunks
    # n should be a positive integer
    def split_progress_interval(self, n):
        prog_start = self.progress[0]
        prog_end = self.progress[1]
        prog_step = (prog_end - prog_start) / n
        result = [None] * n
        for i in range(n):
            result[i] = [prog_start + i * prog_step, prog_start + (i + 1) * prog_step]
        return result

    def get_defn(self):
        return self.system.lookup_defn(self.get_fid())

    # If instance variable is callable, then return var(context_piece)
    # Otherwise, return var
    # This allows things like randomisation to be carried out
    # each time the getter is called
    # The context piece will be the outer piece using this piece to iterate
    def _get_instance_var(self, arg, context_piece=None):
        result = arg
        if callable(result):
            # Evaluate the function on the context, i.e. this fractal piece
            result = result(context_piece)
        return result

    def get_fid(self, context_piece=None):
        return self._get_instance_var(self.fid, context_piece)

    def get_vect(self, context_piece=None):
        return self._get_instance_var(self.vect, context_piece)

    def get_mx(self, context_piece=None):
        return self._get_instance_var(self.mx, context_piece)

    def get_minimum_diameter(self):
        return self.get_defn().get_piece_minimum_diameter(self)

    def plot(self, drawing):
        self.get_defn().plot(drawing, self)

    def iterate(self, collect_next_iteration):
        was_iterated = False
        this_defn = self.get_defn()
        if this_defn is None:
            # If defn cannot be found, this piece should not be added to collect_next_iteration
            pass
        else:
            if not this_defn.should_piece_iterate(self):
                # Do not iterate, but keep piece inside the list collect_next_iteration
                collect_next_iteration.append(self)
            else:
                was_iterated = True
                # 'This' piece (self) is concrete, so fid, vect, mx ought to be values, not functions.
                # Use getters anyway, but don't supply context
                this_vect = self.get_vect()
                this_mx = self.get_mx()
                # defn.children might be a function. Use the getter with this piece (self) as context.
                defn_child_pieces = this_defn.get_children(self)
                count_children = len(defn_child_pieces)
                if 0 < count_children:
                    progress_intervals = self.split_progress_interval(count_children)
                    for i in range(count_children):
                        defn_child_piece = defn_child_pieces[i]
                        # 'Defn child' piece on a fractal definition is abstract,
                        # so fid, vect, mx could be values or functions.
                        # Therefore need to use the getters, evaluated in the context of this piece (self).
                        defn_child_fid = defn_child_piece.get_fid(self)
                        defn_child_vect = defn_child_piece.get_vect(self)
                        defn_child_mx = defn_child_piece.get_mx(self)
                        # 'Next' piece will be concrete, and will be constructed with values (not functions) for fid, vect, mx.
                        next_fid = defn_child_fid
                        next_vect = this_vect + this_mx @ defn_child_vect
                        next_mx = this_mx @ defn_child_mx
                        next_progress = progress_intervals[i]
                        if defn_child_piece.reverse_progress:
                            next_progress = [next_progress[1], next_progress[0]]
                        if defn_child_piece.reset_progress:
                            next_progress = [0, 1]
                        next_piece = FractalPiece(system=self.system, fid=next_fid, vect=next_vect, mx=next_mx, iteration=self.iteration + 1, progress=next_progress)
                        collect_next_iteration.append(next_piece)
        return was_iterated

    def __repr__(self):
        fid = "function" if callable(self.fid) else self.fid
        vect = "function" if callable(self.vect) else self.vect
        mx = "function" if callable(self.mx) else self.mx
        mxprint = f"{mx}".replace("\n", ",")
        return f"(fid {fid}, vect {vect}, mx {mxprint})"



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyautonifty",
    version="0.0.1",
    author="Markichu, Davidryan59, EL-S",
    author_email="EL.S13337@gmail.com",
    description="Automatically and programmatically generate drawing instructions for Nifty.Ink",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Markichu/PythonAutoNifty",
    project_urls={
        "Bug Tracker": "https://github.com/Markichu/PythonAutoNifty/issues",
    },
    keywords = ["nifty.ink", "NFT", "drawing"],
    install_requires=[
        "fonttools==4.28.2",
        "numpy==1.21.4",
        "pygame==2.1.0",
        "scipy==1.7.2",
        "pillow==8.4.0",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "."},
    packages=setuptools.find_packages(where="."),
    python_requires=">=3.6",
)


import numpy as np


# ---------------------------------------
# General helper methods such as easily generate a vector or identity matrix

# Get dimension of vector or matrix (assuming matrix is square)
def np_dim(np_obj):
    return np_obj.shape[0]


# Easy syntax for generating 2D or 3D vectors
def vect(x=0, y=0, z=None, scale=1):
    if z is None:
        return np.array((x, y)) * scale
    return np.array((x, y, z)) * scale


# Find length of vector
# Default power = 2 has 'unit circle' as an actual circle, e.g. Euclidean geometry / Pythagoras' theorem
# Other powers give different shapes of unit circle,
# e.g. power = 0.5 gives a 4 pointed star, 1 gives a diamond, 4 for squircle, 10 is nearly a square
def vect_len(vect, power=2):
    return np.sum(abs(np.float_power(abs(vect), power))) ** (1 / power)


# Find a metric for matrix using length of transformation of x-coord (1, 0)
# Currently 2D only
def metric_matrix_x_coord(mx):
    return (mx[0][0] ** 2 + mx[1][0] ** 2) ** 0.5


# Find a metric for matrix using square root of sum of squares of matrix entries
# Identity matrix in 2D or 3D has metric 1
def metric_matrix_rms(mx):
    return np.sum(mx * mx * (1 / np_dim(mx))) ** 0.5


# Construct metric for matrix using absolute values of the eigenvalues
# Identity matrix in 2D or 3D has metric 1
# Metric using minimum
def metric_matrix_min_eig_val(mx):
    return min(abs(np.linalg.eig(mx)[0]))


# Metric using maximum
def metric_matrix_max_eig_val(mx):
    return max(abs(np.linalg.eig(mx)[0]))


# Metric using ratio of max/min
def metric_matrix_ratio_eig_val(mx):
    result = abs(np.linalg.eig(mx)[0])  # Save the intermediate calculation
    return max(result) / min(result)


# Angle calculator for matrices in O(2) (symmetries of a 2D circle)
# Find angle of vect(1, 0) under transformation by mx
# Returns value between -90 and 270
def mx_angle(mx):
    x = mx[0][0]
    y = mx[0][1]
    deg_from_sign = 180 if x < 0 else 0
    deg_from_angle = 90 if y == 0 else np.degrees(np.arctan(x / y))
    return deg_from_sign + deg_from_angle


# Easy syntax for generating matrix identity
# Defaults to 2D identity matrix
def mx_id(dim=2):
    return np.identity(dim)


# Scale matrix, which is mx_id * scale
# Defaults to 2D identity matrix
def mx_scale(scale=1, dim=2):
    return mx_id(dim) * scale


# Diagonal matrix in 2D or 3D
# Defaults to 2D identity matrix
def mx_diag(x=1, y=1, z=None):
    if z is None:
        return np.diag((x, y))
    return np.diag((x, y, z))


# ---------------------------------------
# Generate rotation matrices in 2D or 3D

# 2D clockwise rotation matrix in XY plane, angle in degrees, +1° rotates Y towards X, with optional reflection
def mx_rotd(angle=0, scale=1, reflect=False):
    angle_in_radians = np.radians(angle)
    c, s = np.cos(angle_in_radians), np.sin(angle_in_radians)
    # TODO: document which plane this is reflecting in
    r = 1
    if reflect:
        r = -1
    return np.array(((c, s), (-s * r, c * r))) * scale


# 3D clockwise rotation matrix in XY plane, angle in degrees, +1° rotates Y towards X
def mx_rotd_XY(angle=0, scale=1):
    angle_in_radians = np.radians(angle)
    c, s = np.cos(angle_in_radians), np.sin(angle_in_radians)
    return np.array(((c, s, 0), (-s, c, 0), (0, 0, 1))) * scale


# 3D clockwise rotation matrix in XZ plane, angle in degrees, +1° rotates Z towards X
def mx_rotd_XZ(angle=0, scale=1):
    angle_in_radians = np.radians(angle)
    c, s = np.cos(angle_in_radians), np.sin(angle_in_radians)
    return np.array(((c, 0, s), (0, 1, 0), (-s, 0, c))) * scale


# 3D clockwise rotation matrix in YZ plane, angle in degrees, +1° rotates Z towards Y
def mx_rotd_YZ(angle=0, scale=1):
    angle_in_radians = np.radians(angle)
    c, s = np.cos(angle_in_radians), np.sin(angle_in_radians)
    return np.array(((1, 0, 0), (0, c, s), (0, -s, c))) * scale


# Utility method to do an XY rotation, then a XZ rotation, then an optional scale
# Applying a matrix transformation is _pre_multiplication, so XZ goes first.
# Scaling is commutative, so can go anywhere.
# Angles are in degrees
def mx_rotd_3D(ang_xy=0, ang_xz=0, scale=1):
    return mx_rotd_XZ(ang_xz) @ mx_rotd_XY(ang_xy, scale)


# ---------------------------------------
# Generate reflection matrices in 2D or 3D

# Reflection matrix in x-axis
def mx_refl_X(dim=2, scale=1):
    if dim == 3:
        return np.array(((-1, 0, 0), (0, 1, 0), (0, 0, 1))) * scale
    return np.array(((-1, 0), (0, 1))) * scale


# Reflection matrix in y-axis
def mx_refl_Y(dim=2, scale=1):
    if dim == 3:
        return np.array(((1, 0, 0), (0, -1, 0), (0, 0, 1))) * scale
    return np.array(((1, 0), (0, -1))) * scale


# Reflection matrix in z-axis
def mx_refl_Z(scale=1):
    return np.array(((1, 0, 0), (0, 1, 0), (0, 0, -1))) * scale


# ---------------------------------------
# Discrete symmetry groups

# Dihedral group of regular polygon 2, 3, 4, 5, 6...
# (rectangle, triangle, square, pentagon, hexagon...)
# where bottom edge is horizontal (so horizontal reflection)
# dh(5, 1) for identity
# dh(5, 2) for 72 degree rotation anticlockwise
# dh(5, 6) for horizontal reflection
# dh(8, 6) for 225 (45 * 5) degree rotation anticlockwise
def mx_dh(sides=3, num=1, scale=1):
    if not isinstance(sides, int):
        raise TypeError("Number of sides must be an integer")
    if sides < 2:
        raise TypeError("Number of sides must be at least 2")
    if not isinstance(num, int):
        raise TypeError("Transformation number must be an integer")
    if num < 1:
        raise TypeError("Transformation number must be at least 1")
    if num > sides * 2:
        raise TypeError(
            f"Transformation number must be at most {sides * 2}")
    rotation_num = (num - 1) % sides  # 0, 1, 2, ..., sides - 1
    reflection_num = ((num - 1) - rotation_num) / sides  # 0, 1
    rotation_mx = mx_rotd((-360 / sides) * rotation_num)
    refl_x_mx = np.array((((-1) ** reflection_num, 0), (0, 1)))
    result = rotation_mx @ refl_x_mx
    return result * scale


# Square transformations 1 to 8
# sq(1) for identity
# sq(2) for 90 degree rotation anticlockwise
# sq(5) for *vertical* reflection (should have been horizontal... but already published on Nifty Ink)
def mx_sq(num=1, scale=1):
    if not isinstance(num, int):
        raise TypeError("Argument is not integer")
    if num < 1:
        raise TypeError("Argument must be at least 1")
    if num > 8:
        raise TypeError("Argument must be at most 8")
    rotation_num = (num - 1) % 4  # 0, 1, 2, 3
    reflection_num = ((num - 1) - rotation_num) / 4  # 0, 1
    rotation_mx = mx_rotd((-90) * rotation_num)
    refl_y_mx = np.array(((1, 0), (0, (-1) ** reflection_num)))
    result = rotation_mx @ refl_y_mx
    return result * scale



from PIL import Image
import random
import math
import numpy as np

from .pos import Pos
from .constants import DRAWING_SIZE, GOLDEN_RATIO, BLACK, WHITE
from .helper_fns import rotate, hsva_to_rgba


# Draw an image from file using dots as pixels
# Many image formats are supported, including .jpg and .png
def point_image(drawing, image_name, do_a_shuffle=False,
                width=DRAWING_SIZE, height=DRAWING_SIZE,
                position=Pos(DRAWING_SIZE / 2, DRAWING_SIZE / 2), keep_aspect_ratio=False, rotation=None):
    # load image from file
    image = Image.open(image_name)

    # convert the image to RGBA values
    rgba_image = image.convert('RGBA')

    # init width and height
    pixel_width, pixel_height = image.size

    if keep_aspect_ratio:  # keep the images original shape
        x_diff = width / max(pixel_width, pixel_height)
        y_diff = height / max(pixel_width, pixel_height)
        max_diff = max(x_diff, y_diff)

        # Centre the image if it isn't a square
        x_offset = abs(max(pixel_width, pixel_height) - pixel_width) / 2
        y_offset = abs(max(pixel_width, pixel_height) - pixel_height) / 2
    else:  # stretch the image to be a square
        x_diff = width / pixel_width
        y_diff = height / pixel_height
        max_diff = max(x_diff, y_diff)

        x_offset = 0
        y_offset = 0

    for x in range(pixel_width):
        for y in range(pixel_height):
            colour = list(rgba_image.getpixel((x, y)))
            colour[3] /= 255

            # Do the initial position calculation
            px = (x + 0.5 + x_offset) * x_diff
            py = (y + 0.5 + y_offset) * y_diff

            # Adjust the position of the image
            px += position.x - (width / 2)
            py += position.y - (height / 2)

            pos = Pos(px, py)

            # Rotate the position
            if rotation:
                pos = rotate(pos, rotation, origin=position)

            pr = max_diff * pow(2, 0.5) / 2
            drawing.add_point(pos=pos, colour=colour, brush_radius=pr)

    if do_a_shuffle:
        drawing.shuffle_lines()

    return drawing


# Draw an image from file using square pixels
def square_image(drawing, image_name, brush_radius=1, do_a_shuffle=False,
                 width=DRAWING_SIZE, height=DRAWING_SIZE,
                 position=Pos(DRAWING_SIZE / 2, DRAWING_SIZE / 2), keep_aspect_ratio=False, rotation=None):
    # load image from file
    image = Image.open(image_name)

    # convert the image to RGBA values
    rgba_image = image.convert('RGBA')

    # init width and height
    pixel_width, pixel_height = image.size

    if keep_aspect_ratio:  # keep the images original shape
        square_pixel_width = width / max(pixel_width, pixel_height)
        square_pixel_height = height / max(pixel_width, pixel_height)

        # Centre the image if it isn't a square
        x_offset = abs(max(pixel_width, pixel_height) - pixel_width) / 2
        y_offset = abs(max(pixel_width, pixel_height) - pixel_height) / 2
    else:  # stretch the image to be a square
        square_pixel_width = width / pixel_width
        square_pixel_height = height / pixel_height

        x_offset = 0
        y_offset = 0

    for x in range(pixel_width):
        for y in range(pixel_height):
            colour = list(rgba_image.getpixel((x, y)))
            colour[3] /= 255

            # Do the initial position calculation
            px = (x + 0.5 + x_offset) * square_pixel_width
            py = (y + 0.5 + y_offset) * square_pixel_height

            # Adjust the position of the image
            px += position.x - (width / 2)
            py += position.y - (height / 2)

            pos = Pos(px, py)

            # Rotate the position
            if rotation:
                pos = rotate(pos, rotation, origin=position)

            drawing.add_rounded_rectangle(centre_pos=pos, width=square_pixel_width, height=square_pixel_height, colour=colour, brush_radius=brush_radius)

    if do_a_shuffle:
        drawing.shuffle_lines()

    return drawing


# Draw four black squares nearly filling the canvas, with different rounding on each corner
def square_example(drawing):
    drawing.add_rounded_square(centre_pos=Pos(250, 250), width=400, colour=BLACK, brush_radius=0.1)
    drawing.add_rounded_square(centre_pos=Pos(750, 250), width=400, colour=BLACK, brush_radius=2)
    drawing.add_rounded_square(centre_pos=Pos(250, 750), width=400, colour=BLACK, brush_radius=40)
    drawing.add_rounded_square(centre_pos=Pos(750, 750), width=400, colour=BLACK, brush_radius=1000)
    return drawing


# Draw four black rectangles nearly filling the canvas, with different rounding on each corner
def rectangle_example(drawing):
    drawing.add_rounded_rectangle(centre_pos=Pos(250, 250), width=400, height=200, colour=BLACK, brush_radius=0.1)
    drawing.add_rounded_rectangle(centre_pos=Pos(750, 250), width=200, height=400, colour=BLACK, brush_radius=2)
    drawing.add_rounded_rectangle(centre_pos=Pos(250, 750), width=400, height=200, colour=BLACK, brush_radius=40)
    drawing.add_rounded_rectangle(centre_pos=Pos(750, 750), width=200, height=400, colour=BLACK, brush_radius=1000)
    return drawing


# Draw a square that appears to rotate on the Nifty Ink canvas
def rotating_square(drawing):
    drawing.add_background((255, 255, 255, 1))
    p1 = Pos(200, 200)
    p2 = Pos(200, 800)
    p3 = Pos(800, 800)
    p4 = Pos(800, 200)
    for _ in range(500):
        drawing.add_line([p1, p2, p3, p4], (0, 0, 0, 1), 10)
        drawing.add_pause(10)
        drawing.add_background((255, 255, 255, 1))
        p1 = rotate(p1, 0.1)
        p2 = rotate(p2, 0.1)
        p3 = rotate(p3, 0.1)
        p4 = rotate(p4, 0.1)
    drawing.add_line([p1, p2, p3, p4], (0, 0, 0, 1), 10)
    return drawing


# Draw a pattern involving a lot of diagonal lines
def tiled_diagonals(drawing, n=50):
    cell_size = DRAWING_SIZE / n
    for x in range(n):
        for y in range(n):
            if random.getrandbits(1):
                pos1 = Pos(x * cell_size, y * cell_size)
                pos2 = Pos((x + 1) * cell_size, (y + 1) * cell_size)
            else:
                pos1 = Pos((x + 1) * cell_size, y * cell_size)
                pos2 = Pos(x * cell_size, (y + 1) * cell_size)
            colour = hsva_to_rgba(random.random(), 0.5, 1)
            drawing.add_straight_line(pos1, pos2, colour, DRAWING_SIZE / n / 10)
    drawing.shuffle_lines()
    return drawing


# Draw a pattern of spiralling dots
def fibonacci_dots(drawing, n=1000):
    for i in range(n):
        # calc radians of rotation and radius from centre
        radians = i * (2 - (2 / GOLDEN_RATIO)) * math.pi
        radius = math.sqrt(i + 1) * DRAWING_SIZE / math.sqrt(n)

        # calc position of dot
        pos = Pos.from_rotational(radians, radius)

        # draw dot
        drawing.add_point(pos, BLACK, 5)

    return drawing


# Draw an image using spiralling dots
def fibonacci_image(drawing, image_filename, n=3000):
    # load image from file
    image = Image.open(image_filename)

    # init result
    width, height = image.size

    for i in range(n):
        # calc radians of rotation and radius from centre
        radians = i * (2 - (2 / GOLDEN_RATIO)) * math.pi
        radius = math.sqrt(i + 1) * DRAWING_SIZE / math.sqrt(n)

        # calc position of dot
        pos = Pos.from_rotational(radians, radius)

        # calc size of dot from image brightness
        image_pos = round((pos.copy() / DRAWING_SIZE) * width) - Pos(1, 1)
        colour = list(image.getpixel((image_pos.x, image_pos.y)))
        brightness = 1 - ((0.2126 * colour[0] + 0.7152 * colour[1] + 0.0722 * colour[2]) / 265)

        # draw dot
        drawing.add_point(pos, colour, brightness * DRAWING_SIZE / math.sqrt(n) / 2)

    return drawing


# Squared Circle example
def squared_circle(drawing, n=8):
    # init step and drawing
    step = 90 / n
    current_h = 0.8

    for i in range(n):
        # get degrees around circle and position
        degrees = i * step + (step / 2)

        # get pos for corners
        pos1 = Pos.from_rotational(np.radians(degrees), DRAWING_SIZE).rotate(np.radians(45))
        pos2 = Pos(pos1.x, DRAWING_SIZE - pos1.y).rotate(np.radians(45))
        pos3 = Pos(DRAWING_SIZE - pos1.x, DRAWING_SIZE - pos1.y).rotate(np.radians(45))
        pos4 = Pos(DRAWING_SIZE - pos1.x, pos1.y).rotate(np.radians(45))

        # do colour
        colour = hsva_to_rgba(current_h, 0.8, 0.85)
        current_h -= 0.7 / n

        # add square
        drawing.add_line([pos1, pos2, pos3, pos4], colour, DRAWING_SIZE / (25 * n))

    return drawing


# Draw straight lines that combine to create impression of curved lines
def curved_lines(drawing, n=20):
    # init drawing
    step = DRAWING_SIZE / n

    # do curved lines
    for i in range(n + 1):
        offset = i * step
        # do bottom left line
        drawing.add_straight_line(Pos(offset, DRAWING_SIZE), Pos(0, offset), BLACK, DRAWING_SIZE / (25 * n))
        # do top right
        drawing.add_straight_line(Pos(DRAWING_SIZE, offset), Pos(offset, 0), BLACK, DRAWING_SIZE / (25 * n))

    # do diagonal lines
    for i in range(n - 1):
        offset = (i + 1) * step
        # do top left line
        drawing.add_straight_line(Pos(offset, 0), Pos(0, offset), BLACK, DRAWING_SIZE / (25 * n))
        # do bottom right
        drawing.add_straight_line(Pos(DRAWING_SIZE, DRAWING_SIZE - offset), Pos(DRAWING_SIZE - offset, DRAWING_SIZE), BLACK, DRAWING_SIZE / (25 * n))

    # do middle diagonal
    drawing.add_straight_line(Pos(DRAWING_SIZE, 0), Pos(0, DRAWING_SIZE), BLACK, DRAWING_SIZE / (25 * n))

    return drawing


# Draw concentric rings of circles getting smaller
def shrinking_circle_ring(drawing, n=20, m=36):
    radians_step = 2 * math.pi / m
    current_h = 0
    big_radius = DRAWING_SIZE / 2
    for _ in range(n):
        # calc current radius
        big_radius = (m * big_radius) / (m + 2 * math.pi)
        small_radius = (math.pi * big_radius) / m

        # do colour
        colour = hsva_to_rgba(current_h, 0.8, 0.9)
        current_h -= 1 / n

        # draw circles around radius
        for i in range(m):
            pos = Pos.from_rotational(radians_step * i, big_radius)
            drawing.add_point(pos, colour, small_radius)
            drawing.add_point(pos, WHITE, small_radius / 3)

    drawing *= 2

    return drawing


# Draw a square fractal which splits a square into 4 more squares (2x2), and iterates on 3 out of 4 smaller squares.
# Note - this is Markichu's original fractal drawing method for square fractals
# It was derived from davidryan59/niftymaestro drawings on Nifty Ink
# Subsequently, davidryan59 implemented more generalised fractal drawing methods, see fractalRunner, FractalSystem etc.
def square_fractal(drawing, master_key, iterations=5):
    # add gradient background
    drawing.add_gradient(Pos(0, 0), Pos(DRAWING_SIZE, DRAWING_SIZE), hsva_to_rgba(0, 0, 0.9), hsva_to_rgba(0, 0, 0.7), 200)
    drawing *= 1 / 0.95

    # Add text
    drawing.write(Pos(600, 125), ["Square", "Fractal", f'#{"".join([str(key) for key in master_key[:3]])}', f"n={iterations}"], 50)

    def rotate_left(ext_key):
        rotate_left_dict = {0: 0,
                            1: 2,
                            2: 3,
                            3: 4,
                            4: 1,
                            5: 6,
                            6: 7,
                            7: 8,
                            8: 5}
        modified_key = [rotate_left_dict[i] for i in ext_key]
        return [modified_key[1], modified_key[3], modified_key[0], modified_key[2]]

    def mirror(ext_key):
        mirror_dict = {0: 0,
                       1: 7,
                       2: 6,
                       3: 5,
                       4: 8,
                       5: 3,
                       6: 2,
                       7: 1,
                       8: 4}
        modified_key = [mirror_dict[i] for i in ext_key]
        return [modified_key[2], modified_key[3], modified_key[0], modified_key[1]]

    def next_ext_key(ext_key, this_key):
        if this_key == 1:
            return ext_key
        elif this_key == 2:
            return rotate_left(ext_key)
        elif this_key == 3:
            return rotate_left(rotate_left(ext_key))
        elif this_key == 4:
            return rotate_left(rotate_left(rotate_left(ext_key)))
        elif this_key == 5:
            return mirror(rotate_left(rotate_left(ext_key)))
        elif this_key == 6:
            return mirror(rotate_left(ext_key))
        elif this_key == 7:
            return mirror(ext_key)
        elif this_key == 8:
            return mirror(rotate_left(rotate_left(rotate_left(ext_key))))

    def square_fractal_recursive(drawing, ext_key, square_pos_corners, iterations):
        pos = (square_pos_corners[0] + square_pos_corners[1]) / 2
        size = (square_pos_corners[1].x - square_pos_corners[0].x) / 2
        if iterations > 0:
            # call recursively for all inner cells
            avg_pos = (square_pos_corners[0] + square_pos_corners[1]) / 2
            subdivisions = [[Pos(square_pos_corners[0].x, avg_pos.y), Pos(avg_pos.x, square_pos_corners[1].y)],
                            [square_pos_corners[0].copy(), avg_pos.copy()],
                            [avg_pos.copy(), square_pos_corners[1].copy()],
                            [Pos(avg_pos.x, square_pos_corners[0].y), Pos(square_pos_corners[1].x, avg_pos.y)]]
            for this_key, subdiv in zip(ext_key, subdivisions):
                if this_key != 0:
                    # draw transparent circle
                    # colour = hsva_to_rgba(0, 0, 0.4 - 0.4 * (pos.y / DRAWING_SIZE), 0.05)
                    # drawing.add_point(pos, colour, size)

                    square_fractal_recursive(drawing, next_ext_key(master_key, this_key), subdiv, iterations - 1)
        else:
            # draw dot in this cell
            colour = hsva_to_rgba(0, 0, 0.4 - 0.4 * (pos.y / DRAWING_SIZE))
            drawing.add_point(pos, colour, size)
        return drawing

    return square_fractal_recursive(drawing, master_key, [Pos(0, 0), Pos(DRAWING_SIZE, DRAWING_SIZE)], iterations)


def text_drawing_example(drawing, font_file_name=None, draw_bounding_box=False):
    drawing.add_gradient(Pos(0, 0), Pos(DRAWING_SIZE, DRAWING_SIZE), hsva_to_rgba(0, 0, 0.9), hsva_to_rgba(0, 0, 0.7), 200)
    drawing.add_gradient(Pos(150, 150), Pos(DRAWING_SIZE - 150, DRAWING_SIZE - 150), hsva_to_rgba(0, 0, 1), hsva_to_rgba(0, 0, 1), 200)

    # Examples of writing:
    # title text
    drawing.write(pos=Pos(160, 170), lines=["Add a title", "here!"], font_size=50, font_file_name=font_file_name, draw_bounding_box=draw_bounding_box)
    # body text
    drawing.write(pos=Pos(160, 300), lines=["Add some interesting",
                                            "content and words",
                                            "here!"], font_size=25, line_spacing=1.35, colour=(64, 64, 200, 1), font_file_name=font_file_name, draw_bounding_box=draw_bounding_box)

    return drawing


def alpha_example(drawing):
    for i in range(random.randint(0, 100)):
        colour = (random.randint(0, 255), random.randint(0, 255), random.randint(0, 255), random.random())
        position = Pos(random.randint(0, DRAWING_SIZE), random.randint(0, DRAWING_SIZE))
        brush_radius = random.randint(1, 500)
        drawing.add_point(position, colour, brush_radius)
        line_points = []
        for j in range(random.randint(2, 20)):
            line_points.append(Pos(random.randint(0, DRAWING_SIZE), random.randint(0, DRAWING_SIZE)))
        colour = (random.randint(0, 255), random.randint(0, 255), random.randint(0, 255), random.random())
        brush_radius = random.randint(1, 10)
        drawing.add_quadratic_bezier_curve(line_points, colour, brush_radius)
    return drawing


def star_example(drawing, lines=10, line_segments=100, draw_points=False, transparency=False, curved_lines=False, enclosed_path=False):
    all_points = []
    drawing.add_point(Pos(DRAWING_SIZE / 2, DRAWING_SIZE / 2), BLACK, 500)
    num = random.randint(2, 10)
    rot = random.random() * 1 + random.randint(0, 1)
    for j in range(random.randint(2, lines)):
        line_points = []
        for i in range(random.randint(2, line_segments)):
            # calc position of dot
            pos = Pos.from_rotational(rot * math.pi * i * j, num * j * i)
            # draw dot
            line_points.append(pos)

        colour = [random.randint(100, 255), random.randint(100, 255), random.randint(100, 255), 1]
        if transparency:
            colour[-1] = random.random()
        brush_radius = 1
        if curved_lines:
            drawing.add_quadratic_bezier_curve(line_points, colour, brush_radius, enclosed_path=True)
        else:
            drawing.add_line(line_points, colour, brush_radius, enclosed_path=enclosed_path)

        all_points.extend(line_points)

    if draw_points:
        for pnt in all_points:
            colour = [random.randint(0, 255), random.randint(0, 255), random.randint(0, 255), 1]
            if transparency:
                colour[-1] = random.random()
            brush_radius = random.randint(10, 20)
            position = pnt + Pos(0.5, 0.5)
            drawing.add_point(position, colour, brush_radius)
        for pnt in all_points:
            colour = [random.randint(0, 255), random.randint(0, 255), random.randint(0, 255), 1]
            if transparency:
                colour[-1] = random.random()
            brush_radius = random.randint(3, 9)
            position = pnt + Pos(0.5, 0.5)
            drawing.add_point(position, colour, brush_radius)
    return drawing



import numpy as np
import random

from .fractal_piece import FractalPiece
from .fractal_system import FractalSystem
from .helper_fns import hex_list_to_rgba
from .constants import DRAWING_SIZE, WHITE, LIGHT_GREY, GREY, DARK_GREY, BLACK, RED, ORANGE, YELLOW, LIGHT_GREEN, GREEN, SPRING_GREEN, CYAN, LIGHT_BLUE, BLUE, PURPLE, MAGENTA, PINK
from .numpy_helper_fns import vect, vect_len, mx_angle, mx_id, mx_scale, mx_diag, mx_rotd, mx_sq
from .fractal_helper_fns import colour_fixed, colour_by_progress, colour_by_tsfm, colour_by_log2_size
from .fractal_helper_fns import plot_dot, plot_path
from .fractal_helper_fns import sort_by_tsfm, grid_generator, wobble_square
from .fractal_helper_fns import get_iteration_fn_standard, get_iteration_fn_stop
from .fractal_generator_fns import gen_children_rand_small_squares, gen_children_fade_out, gen_fid_rand
from .fractal_generator_fns import gen_vect_rand, gen_mx_rand_sq, gen_mx_rand_circ


def fractalRunner(drawing):
    # Use this file to set up a FractalSystem
    # which contains a list of FractalDefn (linked fractals)
    # which each contain a list of FractalPiece (controlling next iteration)
    # as well as a FractalPlotter to control how each fractal definition is drawn

    # ----------------------
    # First, create a fractal system and its definitions
    max_pieces = 100000  # Stop iterating after this number of fractal pieces calculated
    number_of_defns = 50  # Total number of fractal definitions in the linked fractal system

    # Create linked fractal system
    fs = FractalSystem(max_pieces=max_pieces)
    # Create the required number of fractal definitions
    fs.make_defns(number_of_defns)

    # ----------------------
    # Then set parameters that are important for how fractal displays

    # Control randomisation by setting a specific seed
    # This allows pseudorandom fractals to be precisely recreated later on
    random_seed = "My fractal 002"  # Any text here, every text will make a different fractal
    random.seed(random_seed)  # Comment this out if you don't want to control the seeding

    # Initialisation of fractal, its fractal id (fid), position (vector), and orientation (matrix)
    init_defn_fid = 2  # between 0 and number_of_defns-1
    margin_factor = 0.98  # if less than 1, leaves a small gap around the outside of canvas
    init_scale = DRAWING_SIZE / 2
    init_vect = vect(
        x=1,
        y=1,
        scale=init_scale
    )
    init_mx = mx_scale(
        dim=2,
        scale=init_scale * margin_factor
    )
    # This is setup for 2-dimensional fractal

    # Choose a sort order for final list of fractal pieces, which affects drawing order.
    # Examples:
    # fs.piece_sorter = sort_randomly()  # Draw in a random order
    sort_len = 100
    sort_pow = 4
    sort_vect = init_vect
    fs.piece_sorter = sort_by_tsfm(tsfm=lambda vect, mx: vect_len(vect=vect - sort_vect, power=sort_pow) / sort_len, rand=True)  # Draw by distance from sort_vect, with sort_len pixel random boundary
    # fs.piece_sorter = sort_by_z()  # Draw from furthest back to furthest forward (3D only)
    # fs.piece_sorter = sort_by_size()  # Draw from largest at back, to smallest at front

    # Control default iteration function in fractal system
    sys_max_iterations = 20  # Maximum iterations
    sys_min_diameter = 30  # px, fractal pieces will stop iterating if they are smaller than this diameter on at least 1 direction

    # --------------------

    # Definition #0 - empty fractal
    fid = 0
    fd = fs.lookup_defn(fid)
    # no children, iterates to nothing
    fp = fd.plotter
    fp.draws = False  # specify to prevent drawing


    # Definition #1 - identity fractal (doesn't change upon iteration)
    fid = 1
    fd = fs.lookup_defn(fid)
    fd.create_child(fid, vect(0, 0), mx_id())  # this is the identity. Not actually used, since iteration prevented.
    fd.iteration_fn = get_iteration_fn_stop()  # specify to prevent further calculation of iterations
    fp = fd.plotter
    fp.draws = False


    # Definition #2 - use as wrapper to display 1 or more other fractals
    # Since this is a wrapper, reset the progress on each inner fractal
    # to reset things like colouring schemes
    fid = 2
    fd = fs.lookup_defn(fid)
    scv = 0.5
    scm = 0.45
    fd.create_child(3, vect(-1, -1) * scv, mx_scale(scm), reset_progress=True)
    fd.create_child(4, vect(1, -1) * scv, mx_scale(scm), reset_progress=True)
    fd.create_child(5, vect(-0.8, 1) * scv, mx_scale(scm * 0.8), reset_progress=True)
    fd.create_child(6, vect(-0.4, -0.6) * scv, mx_scale(0.45 * scm), reset_progress=True)
    fd.create_child(7, vect(-1.5, 1.6) * scv, mx_rotd(angle=-60, scale=0.35 * scm), reset_progress=True)
    fd.create_child(8, vect(1, 1) * scv, mx_scale(scm), reset_progress=True)
    fd.create_child(9, vect(-1, 0.2) * scv, mx_rotd(angle=20, scale=scm * 1.15) @ np.array(((1, 0), (0, -1))), reset_progress=True)
    # keep default plotting setup, this definition likely won't plot
    # since it iterates to other things


    # Definition #3 - demo Sierpinski Sieve
    fid = 3
    fd = fs.lookup_defn(fid)
    n = 2
    sc = 1 / n
    grid = grid_generator(x_steps=n, y_steps=n)  # Defaults from -1 to +1 in both x and y directions
    # Grid generator generates coordinates at square midpoints from -1 to 1, total steps 2
    # grid(0, 0) = (-0.5, -0.5)
    # grid(1, 0) = (0.5, -0.5)
    # etc
    fd.create_child(fid, grid(0, 0), mx_id() * sc)  # Same scale matrix result, three ways of writing
    fd.create_child(fid, grid(0.05, 1), mx_scale(sc))
    fd.create_child(fid, grid(1, 0.05), mx_sq(num=1, scale=sc))

    # Showing here a manual assignment of relative diameter to fractal definition.
    # This is redundant since 2 is the default value, but larger or smaller values will affect drawing and iteration.
    # Min diameter 2 means in some orientation the defn fits between planes 2 units apart, for a piece with identity transformation (vect(0, 0), mx_id())
    fd.relative_diameter = 2
    fd.iteration_fn = get_iteration_fn_standard(min_diameter=60, max_iterations=4)

    fp = fd.plotter
    x_minus_y = lambda vect, mx: vect[0] - vect[1]
    fill_colour_fn = colour_by_tsfm(-150, 150, tsfm=x_minus_y, colours=[RED, BLUE], alpha=0.5)
    outline_colour_fn = colour_fixed(colour=BLACK, alpha=0.75)
    # Plotting method: uses convex hull on definition by default, override by specifying vector_list = [vect(x, y)...]
    # Setting shrink=True (overriding default value of False) moves the shape's vectors inwards
    # to try and draw strictly inside the original shape using the specified brush radius (width).
    # If shape is too small, compared with brush radius, then the lines drawn will overlap the boundary anyway.
    fill_plot_fn = plot_path(width=5, expand_factor=1.00, fill=True, shrink=True)
    outline_plot_fn = plot_path(width=3, fill=False, closed=True, shrink=False)
    # Composite drawing method: do a translucent fill first, then a nearly solid outline
    # The fill uses the shrink method, the outline does not, which gives some interesting effects at the boundaries.
    fp.add(fill_plot_fn, fill_colour_fn)
    fp.add(outline_plot_fn, outline_colour_fn)


    # Definition #4 - demo of random square matrix transformations
    fid = 4
    fd = fs.lookup_defn(fid)
    n = 2
    sc = 1 / n
    grid = grid_generator(x_steps=n, y_steps=n)
    fd.create_child(fid, grid(0, 0), mx_scale(sc))
    fd.create_child(fid, grid(1, 0), mx_scale(sc))
    fd.create_child(fid, grid(0, 1), gen_mx_rand_sq(scale=sc ** 1.3))
    fd.create_child(fid, grid(1, 1), gen_mx_rand_sq(scale=sc ** 1.7))
    fp = fd.plotter
    colouring_fn = colour_by_log2_size(2, 4, colours=[GREEN, BLUE])
    plotting_fn = plot_path(
        closed=True,
        width=2,
        expand_factor=1.1,
        wobble_fn=wobble_square(pixels=3),
        vector_list=[vect(-1, 0), vect(-1, -1), vect(1, -1), vect(1, 0), vect(0, 1)],
        curved=True
    )
    fp.add(plotting_fn, colouring_fn)


    # Definition #5 - demo of random hexagon fractal
    fid_exit = 1
    fid_iterate = 5
    fd = fs.lookup_defn(fid_iterate)
    sc = 1 / 3
    h = 3 ** 0.5
    id_list = [fid_exit, fid_iterate, fid_iterate, fid_iterate, fid_iterate]
    fid_fn = gen_fid_rand(id_list)
    mx_fn = gen_mx_rand_circ(sc)
    fd.create_child(fid_fn, vect(2, 0) * sc, mx_fn)
    fd.create_child(fid_fn, vect(1, h) * sc, mx_fn)
    fd.create_child(fid_fn, vect(-1, h) * sc, mx_fn)
    fd.create_child(fid_fn, vect(-2, 0) * sc, mx_fn)
    fd.create_child(fid_fn, vect(-1, -h) * sc, mx_fn)
    fd.create_child(fid_fn, vect(1, -h) * sc, mx_fn)
    fd.create_child(fid_fn, vect(0, 0) * sc, mx_fn)
    fp = fd.plotter
    colouring_fn = colour_by_progress(colours=[BLACK, PINK, LIGHT_BLUE, GREEN, YELLOW, BLACK])
    # plotting_fn = plot_dot(expand_factor=0.5)  # expand_factor < 1 makes dots distinct
    plotting_fn = plot_path(expand_factor=0.5, fill=True)
    fp.add(plotting_fn, colouring_fn)


    # Definition #6 - demo of random vector shift
    fid = 6
    fd = fs.lookup_defn(fid)
    sc = 0.5
    fd.create_child(fid, vect(-1, -1) * sc, mx_scale(sc))  # Two versions of same scale matrix here
    fd.create_child(fid, vect(1, -1) * sc, mx_diag(sc, sc))
    fd.create_child(fid, gen_vect_rand([-sc, sc], [sc, sc]), mx_scale(sc))
    fp = fd.plotter
    distance_from_origin = lambda vect, mx: ((vect[0]) ** 2 + (vect[1]) ** 2) ** 0.5
    colouring_fn = colour_by_tsfm(500, 580, tsfm=distance_from_origin, colours=[PURPLE, YELLOW, PINK])
    plotting_fn = plot_dot(expand_factor=1.5, wobble_fn=wobble_square(pixels=5))  # make dots overlap
    fp.add(plotting_fn, colouring_fn)


    # Definition #7 - Standard (2D) dragon curve
    fid = 7
    fd = fs.lookup_defn(fid)
    sc = 0.5 ** 0.5
    fd.create_child(fid, vect(-0.5, 0.5), mx_rotd(angle=-45, scale=sc))
    fd.create_child(fid, vect(0.5, 0.5), mx_rotd(angle=-135, scale=sc))
    fp = fd.plotter
    piece_angle = lambda vect, mx: mx_angle(mx)
    colouring_fn = colour_by_tsfm(-90, 270, tsfm=piece_angle, colours=[RED, YELLOW, GREEN, BLUE])
    plotting_fn = plot_path(width=3, vector_list=[vect(-1, 0), vect(1, 0)])
    fp.add(plotting_fn, colouring_fn)


    # Definition #8 - Random Sierpinski Carpet
    # gen_children_rand_small_squares - 7 out of 9 little squares, in 3x3 big square
    # gen_children_fade_out - Iterate with a probability depending on how far fractal piece is from a specified point
    fidr = 8
    fd = fs.lookup_defn(fidr)
    # fd.children = gen_children_rand_small_squares(system=fs, fid=fidr, m=7, n=3)  # Definition children is dynamically calculated by callback
    fd.children = gen_children_fade_out(system=fs, fid=fidr, n=3, centre_vect=vect(750, 750), cutoff_diameter=50, d1=75, d2=225, p1=1, p2=0)
    fp = fd.plotter
    x, y = 650, 650
    distance_from_pt = lambda vect, mx: ((vect[0] - x) ** 2 + (vect[1] - y) ** 2) ** 0.5
    colouring_fn = colour_by_tsfm(50, 350, tsfm=distance_from_pt, colours=[MAGENTA, YELLOW, CYAN])
    plotting_fn = plot_path(fill=True, width=3, expand_factor=0.6, vector_list=[vect(-1, 1), vect(-1, -1), vect(1, -1), vect(1, 0), vect(0, 0)])
    fp.add(plotting_fn, colouring_fn)


    # Definition #9 - Modified dragon curve with fractal dimension varying from 1 to 2, from one end to the other
    def gen_children_variable_dragon(system, fid):
        def calc_children(context_piece=None):
            progress = 1
            if context_piece is not None:
                progress = context_piece.get_progress_value()
            x = 0
            y = progress
            x1 = 0.5 * (-1 + x)
            x2 = 0.5 * (1 + x)
            y1 = 0.5 * y
            y2 = 0.5 * y
            piece1 = FractalPiece(system=system, fid=fid, vect=vect(x1, y1), mx=np.array(((x - x1, y1 - y), (y - y1, x - x1))))
            piece2 = FractalPiece(system=system, fid=fid, vect=vect(x2, y2), mx=np.array(((x - x2, y2 - y), (y - y2, x - x2))))
            piece2.reverse_progress = True  # Needed to get linear colouring and iteration! Try removing it...
            children = [piece1, piece2]
            return children

        return calc_children

    fid = 9
    fd = fs.lookup_defn(fid)
    sc = 0.5 ** 0.5
    fd.children = gen_children_variable_dragon(fs, fid)
    fd.iteration_fn = get_iteration_fn_standard(min_diameter=20, max_iterations=20)
    fp = fd.plotter
    colouring_fn = colour_by_progress([BLUE, RED, ORANGE])
    # colouring_fn = colour_fixed(CYAN, alpha=0.75) # Alternative fixed colouring method
    plotting_fn = plot_path(width=2, vector_list=[vect(-1, 0), vect(1, 0)])
    # plotting_fn = plot_path(width=1, closed=True, vector_list=[vect(-1, 0), vect(0, -1), vect(1, 0), vect(0, 1)])  # Alternative plot method
    fp.add(plotting_fn, colouring_fn)


    # --------------------

    # Default function to control when pieces stop iterating
    # This can be overwritten at the fractal definition level
    fs.iteration_fn = get_iteration_fn_standard(min_diameter=sys_min_diameter, max_iterations=sys_max_iterations)

    # --------------------

    # TODO: Convex Hulls only currently works for 2D. Can it work for 3D too?
    # Calculate Convex Hulls after fractal definitions completed
    # If this is not done, plot_path will not have hulls available for drawing,
    # and will fall back to unit squares if a path (vector_list) is not supplied
    fs.calculate_hulls(hull_accuracy=0.1, max_iterations=10)

    # --------------------

    # # Examine the fractal system setup
    # fs.log("")
    # fs.log("Fractal System:")
    # fs.log(fs)
    # fs.log("")

    # Calculate the iterations
    initial_piece = FractalPiece(system=fs, fid=init_defn_fid, vect=init_vect, mx=init_mx)
    fs.initial_pieces = [initial_piece]
    fs.do_iterations()
    fs.log(f"After iteration, there are {fs.final_size()} pieces")
    fs.log("")

    # Plot iterations to drawing
    fs.plot(drawing)
    fs.log("Fractal System plotted successfully")
    fs.log("")

    return drawing
